#include <stdio.h>
#include <stdlib.h>
#include <algorithm>
#include "\LTR\include\ltrapi_ll.h"
#include "USB_defs.h"
#include "LUSB.h"

#define RELAXED_FPGAINFO_CHECK  1

typedef enum
{
        V_RESET_FPGA        = 0,
        V_PUT_ARRAY         = 1,
        V_GET_ARRAY         = 2,
        V_INIT_FPGA         = 3,
        V_SET_DMA_PARAMETRS = 4,
        V_GET_MODULE_NAME   = 11,
        V_SET_WAIT_TIME     = 12,
        V_CALL_APPLICATION  = 15
}VendorCommand;

#pragma pack(1)
typedef struct
{
    BYTE Name[16];
    BYTE SerialNumber[16];
    BYTE CpuType[16];
    BYTE Revision;
    long QuartzFrequency;
    BYTE Unused[128-53];    
}_DEVICE_DESCR;
#pragma pack()

//-----------------------------------------------------------------------------
// TLTR010_BOOT_LOADER
//-----------------------------------------------------------------------------
// 
//-----------------------------------------------------------------------------
TLTR010_BOOT_LOADER::TLTR010_BOOT_LOADER() 
{
}
TLTR010_BOOT_LOADER::~TLTR010_BOOT_LOADER()
{
}

bool WINAPI TLTR010_BOOT_LOADER::SetUSBDevice(HANDLE Device)
{
    this->MainDevice=Device;
    return true;
}

//-----------------------------------------------------------------------------
//
//-----------------------------------------------------------------------------
bool WINAPI TLTR010_BOOT_LOADER::CallApplication(DWORD address)
{
    _ControlRequest Request;
    Request.RequestDirectionIN=false;
	Request.bRequest=V_CALL_APPLICATION;
	Request.wValue=address;
	Request.wIndex=(address)>>16;

    TLUSB_BASE * Device = (TLUSB_BASE *)MainDevice;
    return Device->CommandControl(&Request);
}

bool WINAPI TLTR010_BOOT_LOADER::GetArray(DWORD address, BYTE *data, DWORD size)
{

 if(data==NULL || size==0) return false;

    _ControlRequest Request;
    Request.RequestDirectionIN=true;
	Request.bRequest=14; // get array Bootloader

 TLUSB_BASE * Device = (TLUSB_BASE *)MainDevice;
 DWORD i=0;
 while(i<size)
 {
        DWORD sz=std::min<int>(4096, size-i);
     	Request.wValue=address;
	    Request.wIndex=(address)>>16;

        if(!Device->ControlLDevice(0,&Request, data+i, sz, 4000)) return false;
        //
        address+=sz;
        i+=sz;
 }

 return true;
}

bool WINAPI TLTR010_BOOT_LOADER::PutArray(DWORD address, BYTE *data, DWORD size)
{

 if(data==NULL || size==0) return false;

    _ControlRequest Request;
    Request.RequestDirectionIN=false;
	Request.bRequest=13; // put array Bootloader

 TLUSB_BASE * Device = (TLUSB_BASE *)MainDevice;
 DWORD i=0;
 while(i<size )
 {
        DWORD sz=std::min<int>(4096, size-i);
     	Request.wValue=address;
	    Request.wIndex=(address)>>16;

        if(!Device->ControlLDevice(0,&Request, data+i, sz, 4000)) return false;
        //
        address+=sz;
        i+=sz;
 }

 return true;
}
//-----------------------------------------------------------------------------


//-----------------------------------------------------------------------------
bool WINAPI TLTR010_BOOT_LOADER::GetModuleDescription(TDESCRIPTION_LTR010 *description)
{
 if(description==NULL) return false;

    BYTE info1[16], info2[16];
    _DEVICE_DESCR ddescr;
    unsigned int retries = 0;

    if (!GetArray(LTR010_MODULE_DESCRIPTOR_ADDRESS, (BYTE*)&ddescr, sizeof(ddescr)))
        return false;
    retries = std::max(retries, ((TLUSB_BASE*)MainDevice)->LastIoctlRetries);

    if (!GetArray(LTR010_BOOT_LOADER_VERSION_ADDRESS, info1, 16))
        return false;
    retries = std::max(retries, ((TLUSB_BASE*)MainDevice)->LastIoctlRetries);

    if (!GetArray(LTR010_FIRMWARE_VERSION_ADDRESS, info2, 16))
        return false;
    retries = std::max(retries, ((TLUSB_BASE*)MainDevice)->LastIoctlRetries);

    ((TLUSB_BASE*)MainDevice)->LastIoctlRetries = retries;

        memset(description, 0, sizeof(TDESCRIPTION_LTR010));
        // module description
        TDESCRIPTION_MODULE &mdescr=description->Module;
        strncpy(mdescr.CompanyName,  "L-CARD Inc.",      sizeof(mdescr.CompanyName));
        strncpy(mdescr.DeviceName,   ddescr.Name,         std::min(sizeof(mdescr.DeviceName), sizeof(ddescr.Name)));
        strncpy(mdescr.SerialNumber, ddescr.SerialNumber, std::min(sizeof(mdescr.SerialNumber), sizeof(ddescr.SerialNumber)));
        strncpy(mdescr.Comment,      ddescr.Unused,       std::min(sizeof(mdescr.Comment), sizeof(ddescr.Unused)));
        //mdescr.Revision=ddescr.DeviceRevision;
        // cpu description
        TDESCRIPTION_CPU &cdescr=description->Cpu;
        cdescr.Active=1;
        strncpy(cdescr.Name, ddescr.CpuType, std::min(sizeof(cdescr.Name), sizeof(ddescr.CpuType)));
        cdescr.ClockRate=ddescr.QuartzFrequency;
        cdescr.FirmwareVersion=*(DWORD*)info2;
        sprintf(cdescr.Comment, "Bootloader v%d.%d.%d.%d", (int)info1[3], (int)info1[2], (int)info1[1], (int)info1[0]);
        // interface description
        TDESCRIPTION_INTERFACE &idescr=description->Interface;
        idescr.Active=1;
        strcpy(idescr.Name, "USB 2.0");
        strcpy(idescr.Comment, "HS/FS Mode");

    return true;
}
//-----------------------------------------------------------------------------
//
//-----------------------------------------------------------------------------
bool WINAPI TLTR010_BOOT_LOADER::SetModuleDescription(TDESCRIPTION_LTR010 *description)
{
 if(description==NULL) return false;

 if (!CallApplication(LTR010_BOOT_LOADER_START_ADDRESS)) return false;

        _DEVICE_DESCR ddescr;
        memset(&ddescr, 0, sizeof(ddescr));
        // module description
        TDESCRIPTION_MODULE &mdescr=description->Module;
        strncpy(ddescr.Name,         mdescr.DeviceName,   std::min(sizeof(mdescr.DeviceName), sizeof(ddescr.Name)));
        strncpy(ddescr.SerialNumber, mdescr.SerialNumber, std::min(sizeof(mdescr.SerialNumber), sizeof(ddescr.SerialNumber)));
        strncpy(ddescr.Unused,       mdescr.Comment,      std::min(sizeof(mdescr.Comment), sizeof(ddescr.Unused)));
        // cpu description
        TDESCRIPTION_CPU &cdescr=description->Cpu;
        strncpy(ddescr.CpuType, cdescr.Name, std::min(sizeof(cdescr.Name), sizeof(ddescr.CpuType)));
        ddescr.QuartzFrequency=cdescr.ClockRate;
        // write descriptor
        PutArray(LTR010_MODULE_DESCRIPTOR_ADDRESS, (BYTE*)&ddescr, sizeof(ddescr));


 return true;
}

//-----------------------------------------------------------------------------
// TLTR010
//-----------------------------------------------------------------------------
// 
//-----------------------------------------------------------------------------
TLTR010::TLTR010()
{
    NoRealLoadFPGAGlag=false;
    memset(fpga_info, 0, sizeof(fpga_info));
}

TLTR010::~TLTR010()
{
}

bool WINAPI TLTR010::SetUSBDevice(HANDLE Device)
{
    this->MainDevice=Device;
    return true;
}

//-----------------------------------------------------------------------------
// 
//-----------------------------------------------------------------------------
bool WINAPI TLTR010::GetArray(DWORD address, BYTE *data, DWORD size)
{

 if(data==NULL || size==0) return false;

    _ControlRequest Request;
    Request.RequestDirectionIN=true;
	Request.bRequest=V_GET_ARRAY;

 TLUSB_BASE * Device = (TLUSB_BASE *)MainDevice;
 DWORD i=0;
 while(i<size )
 {
        DWORD sz=std::min<int>(4096, size-i);
     	Request.wValue=address;
	    Request.wIndex=(address)>>16;

        if(!Device->ControlLDevice(0,&Request, data+i, sz, 4000)) return false;
        //
        address+=sz;
        i+=sz;
 }

 return true;
}

bool WINAPI TLTR010::PutArray(DWORD address, BYTE *data, DWORD size)
{

 if(data==NULL || size==0) return false;

    _ControlRequest Request;
    Request.RequestDirectionIN=false;
	Request.bRequest=V_PUT_ARRAY;

 TLUSB_BASE * Device = (TLUSB_BASE *)MainDevice;
 DWORD i=0;
 while(i<size )
 {
        DWORD sz=std::min<int>(4096, size-i);
     	Request.wValue=address;
	    Request.wIndex=(address)>>16;

        if(!Device->ControlLDevice(0,&Request, data+i, sz, 4000)) return false;
        //
        address+=sz;
        i+=sz;
 }

 return true;
}
//------------------------------------------------------------------------
//   
//------------------------------------------------------------------------
bool WINAPI TLTR010::ResetFPGA(void)
{
     _ControlRequest Request;
    Request.RequestDirectionIN=false;
	Request.bRequest=V_RESET_FPGA;
	Request.wValue=0;
	Request.wIndex=0;

    TLUSB_BASE * Device = (TLUSB_BASE *)MainDevice;
    return Device->CommandControl(&Request);
}
//------------------------------------------------------------------------------
//   
//------------------------------------------------------------------------------
bool WINAPI TLTR010::InitFPGA(void)
{
    _ControlRequest Request;
    Request.RequestDirectionIN=false;
	Request.bRequest=V_INIT_FPGA;
	Request.wValue=0;
	Request.wIndex=0;
    TLUSB_BASE * Device = (TLUSB_BASE *)MainDevice;
    return Device->CommandControl(&Request);
}

//------------------------------------------------------------------------------
//   ,    100   17  (1..170)
//------------------------------------------------------------------------------
bool WINAPI TLTR010::SetWaitTime(WORD mks100)
{
    _ControlRequest Request;
    Request.RequestDirectionIN=false;
	Request.bRequest=V_SET_WAIT_TIME;
	Request.wValue=mks100;
	Request.wIndex=0;
    TLUSB_BASE * Device = (TLUSB_BASE *)MainDevice;
    return Device->CommandControl(&Request);
}

//------------------------------------------------------------------------------
//   dmac
//------------------------------------------------------------------------------
// RDMA_PARAM
// 0 - disable rdma
// 1 - start rdma if 0..1K
// 2 - start rdma if 1K..4K
// 3 - start rdma if 0..16K
// 4 - start rdma if 16K..
// 5 - start rdma if 16K..
// 6 - start rdma if 1K..
// 7 - start rdma if 0..
// WDMA_PARAM
// 0 - buffer length
// 1 - (     256)
// 2 - (  1*1024)
// 3 - (  4*1024)
// 4 - ( 16*1024)
// 5 - ( 64*1024)
// 6 - (128*1024)
// 7 - (256*1024)
//------------------------------------------------------------------------------
bool WINAPI TLTR010::InitDMAC(BYTE RDMA_PARAM, BYTE WDMA_PARAM)
{
    _ControlRequest Request;
    Request.RequestDirectionIN=false;
	Request.bRequest=V_SET_DMA_PARAMETRS;
	Request.wValue=(RDMA_PARAM<<8)|WDMA_PARAM;
	Request.wIndex=3;
    TLUSB_BASE * Device = (TLUSB_BASE *)MainDevice;
    return Device->CommandControl(&Request);
}
//------------------------------------------------------------------------------
//   fpga
//------------------------------------------------------------------------------
bool WINAPI TLTR010::LoadFPGA(BYTE *data, BYTE RDMA_PARAM, BYTE WDMA_PARAM, LPINT error_code)
    {
    INT err = 0;
    unsigned int retries = 0;

#define CHECK(call, ecode)      \
    if (!(call)) { err = (ecode); break; }  \
    retries = std::max(retries, ((TLUSB_BASE*)MainDevice)->LastIoctlRetries)

    do
        {
        if (!data)              { err = LTR_FPGA_ERR_BADPARAM; break; }
        memset(fpga_info, 0, sizeof(fpga_info));
        CHECK(InitDMAC(0,0),    LTR_FPGA_ERR_INITDMAC0);
        CHECK(ResetFPGA(),      LTR_FPGA_ERR_RESETFPGA);
        Sleep(100);
        if (!NoRealLoadFPGAGlag)
            {
            CHECK(PutArray(SEL_FPGA_DATA, data, EP1K50_SIZE), LTR_FPGA_ERR_PUTARRAY);
            }
        CHECK(InitFPGA(),       LTR_FPGA_ERR_INITFPGA);
        CHECK(InitDMAC(7, 0),   LTR_FPGA_ERR_INITDMAC7);

        WORD fpga_info_buf[512];
        DWORD bytes_read = 0;

        Sleep(1000);
        if (!((TLUSB_BASE*)MainDevice)->ReadBulk((BYTE*)fpga_info_buf, sizeof(fpga_info_buf), &bytes_read, 5000))
            { err = LTR_FPGA_ERR_READBULK; break; }
        if (!bytes_read)
            { err = LTR_FPGA_ERR_NOREPLY; break; }

#if RELAXED_FPGAINFO_CHECK
        /* 17-Jun-2013:    ,    LTR-U-1,    
              (,       ) */
        int i, word_count = bytes_read / sizeof(fpga_info_buf[0]);
        for (i = 0; i < word_count; i++)
            {
            WORD w = fpga_info_buf[i];
            if (!w)
                { /* { 0, 0 } =  , {0, x!=0 } =  */
                if ((i < word_count - 1) && fpga_info_buf[i + 1])
                    err = LTR_FPGA_ERR_BADREPLY;
                break;
                }
            if ((w & 0xFF00) != 0xE000)
                {
                err = LTR_FPGA_ERR_BADREPLY;
                break;
                }
            if (i < (int)sizeof(fpga_info) - 1) /*   fpga_info   0,   */
                fpga_info[i] = w & 0xFF;
            }
#else
        /*       */
        int nzw_count = bytes_read / sizeof(fpga_info_buf[0]);
        while ((nzw_count > 0) && !fpga_info_buf[nzw_count - 1])
            nzw_count--;

        for (int i = 0; i < nzw_count; i++)
            {
            WORD w = fpga_info_buf[i];
            if ((w & 0xFF00) != 0xE000)
                { err = LTR_FPGA_ERR_BADREPLY; break; }
            if (i < (int)sizeof(fpga_info) - 1) /*   fpga_info   0,   */
                fpga_info[i] = w & 0xFF;
            }
#endif

        if (err) break;

        CHECK(InitDMAC(RDMA_PARAM, WDMA_PARAM), LTR_FPGA_ERR_INITDMAC);
        }
    while (0);

    if (error_code) *error_code = (err) ? err : retries;
    return !err;
#undef CHECK
    }
//-----------------------------------------------------------------------------
bool WINAPI TLTR010::LoadFPGAFromFile(PCHAR fname, BYTE RDMA_PARAM, BYTE WDMA_PARAM, LPINT error_code)
    {
    BYTE data[EP1K50_SIZE];
    INT err = 0;
    memset(data, 0, sizeof(data));

    if (!NoRealLoadFPGAGlag)
        {
        if (!fname || !fname[0])
            {
            err = LTR_FPGA_ERR_BADPARAM;
            }
        else
            {
            FILE *file = fopen(fname, "rt");
            if (file)
                {
                DWORD len = 0;
                while (!feof(file) && (len < sizeof(data)))
                    {
                    int d;
                    if (fscanf(file, "%3d,", &d) < 0)
                        {
                        err = LTR_FPGA_ERR_FILEIO;
                        break;
                        }
                    data[len++] = d;
                    }
                fclose(file);
                if (!len) //   
                    err = LTR_FPGA_ERR_FILEIO;
                }
            else
                {
                err = LTR_FPGA_ERR_FILEIO;
                }
            }
        }

    if (error_code) *error_code = err;
    if (err) return false;
    return LoadFPGA(data, RDMA_PARAM, WDMA_PARAM, error_code);
    }
//------------------------------------------------------------------------------
//
//------------------------------------------------------------------------------
bool WINAPI TLTR010::GetModuleDescription(TDESCRIPTION_LTR010 *description)
{
    if(description==NULL) return false;

    BYTE info1[16], info2[16];
    _DEVICE_DESCR ddescr;
    unsigned int retries = 0;

    if (!GetArray(LTR010_MODULE_DESCRIPTOR_ADDRESS, (BYTE*)&ddescr, sizeof(ddescr)))
        return false;
    retries = std::max(retries, ((TLUSB_BASE*)MainDevice)->LastIoctlRetries);

    if (!GetArray(LTR010_BOOT_LOADER_VERSION_ADDRESS, info1, 16))
        return false;
    retries = std::max(retries, ((TLUSB_BASE*)MainDevice)->LastIoctlRetries);

    if (!GetArray(LTR010_FIRMWARE_VERSION_ADDRESS, info2, 16))
        return false;
    retries = std::max(retries, ((TLUSB_BASE*)MainDevice)->LastIoctlRetries);

    ((TLUSB_BASE*)MainDevice)->LastIoctlRetries = retries;

        memset(description, 0, sizeof(TDESCRIPTION_LTR010));
        // module description
        TDESCRIPTION_MODULE &mdescr=description->Module;
        strncpy(mdescr.CompanyName,  "L-CARD Inc.",      sizeof(mdescr.CompanyName));
        strncpy(mdescr.DeviceName,   ddescr.Name,         std::min(sizeof(mdescr.DeviceName), sizeof(ddescr.Name)));
        strncpy(mdescr.SerialNumber, ddescr.SerialNumber, std::min(sizeof(mdescr.SerialNumber), sizeof(ddescr.SerialNumber)));
        strncpy(mdescr.Comment,      ddescr.Unused,       std::min(sizeof(mdescr.Comment), sizeof(ddescr.Unused)));
        //mdescr.Revision=ddescr.DeviceRevision;
        // cpu description
        TDESCRIPTION_CPU &cdescr=description->Cpu;
        cdescr.Active=1;
        strncpy(cdescr.Name, ddescr.CpuType, std::min(sizeof(cdescr.Name), sizeof(ddescr.CpuType)));
        cdescr.ClockRate=ddescr.QuartzFrequency;
        cdescr.FirmwareVersion=*(DWORD*)info2;
        sprintf(cdescr.Comment, "Bootloader v%d.%d.%d.%d", (int)info1[3], (int)info1[2], (int)info1[1], (int)info1[0]);
        // interface description
        TDESCRIPTION_INTERFACE &idescr=description->Interface;
        idescr.Active=1;
        strcpy(idescr.Name, "USB 2.0");
        strcpy(idescr.Comment, "HS/FS Mode");

    // fpga description
    if(fpga_info[0])
    {
        TDESCRIPTION_FPGA &fdescr=description->Fpga;
        fdescr.Active=1;
        fdescr.ClockRate=60e6;
        // Name
        char *ptag, *p1, *p2;
        ptag=strstr((char*)fpga_info, "<Name>");
        if(ptag && (p1=strchr(ptag, '{'))!=0 && (p2=strchr(ptag, '}'))!=0)
        {
            strncpy(fdescr.Name, p1+1, std::min<int>(sizeof(fdescr.Name), p2-p1-1));
        }
        // Version
        ptag=strstr((char*)fpga_info, "<Version>");
        if(ptag && (p1=strchr(ptag, '{'))!=0 && (p2=strchr(ptag, '}'))!=0)
        {

            int i, j, k;
            if(sscanf(p1, "{%d.%d.%d", &i, &j, &k)==3)
            {
                fdescr.FirmwareVersion=((i&255)<<24)|((j&255)<<16)|((k&255)<<8);
            }

        }
        // Comment
        ptag=strstr((char*)fpga_info, "<Comments>");
        if(ptag && (p1=strchr(ptag, '{'))!=0 && (p2=strchr(ptag, '}'))!=0)
        {
            strncpy(fdescr.Comment, p1+1, std::min<int>(sizeof(fdescr.Comment), p2-p1-1));
        }
    }

 return true;
}

//------------------------------------------------------------------------------
//
//------------------------------------------------------------------------------
bool WINAPI TLTR010::ConvertFPGADescription(TDESCRIPTION_LTR010 *description)
{
    if(description==NULL) return false;

    // fpga description
    if(fpga_info[0])
    {
        TDESCRIPTION_FPGA &fdescr=description->Fpga;
        fdescr.Active=1;
        fdescr.ClockRate=60e6;
        // Name
        char *ptag, *p1, *p2;
        ptag=strstr((char*)fpga_info, "<Name>");
        if(ptag && (p1=strchr(ptag, '{'))!=0 && (p2=strchr(ptag, '}'))!=0)
        {
            strncpy(fdescr.Name, p1+1, std::min<int>(sizeof(fdescr.Name), p2-p1-1));
        }
        // Version
        ptag=strstr((char*)fpga_info, "<Version>");
        if(ptag && (p1=strchr(ptag, '{'))!=0 && (p2=strchr(ptag, '}'))!=0)
        {

            int i, j, k;
            if(sscanf(p1, "{%d.%d.%d", &i, &j, &k)==3)
            {
                fdescr.FirmwareVersion=((i&255)<<24)|((j&255)<<16)|((k&255)<<8);
            }

        }
        // Comment
        ptag=strstr((char*)fpga_info, "<Comments>");
        if(ptag && (p1=strchr(ptag, '{'))!=0 && (p2=strchr(ptag, '}'))!=0)
        {
            strncpy(fdescr.Comment, p1+1, std::min<int>(sizeof(fdescr.Comment), p2-p1-1));
        }
    }

    return true;
}




