#ifndef __USBAPI__
#define __USBAPI__
// -----------------------------------------------------------------------------
// ---------------------------- COMMON PART ------------------------------------
// -----------------------------------------------------------------------------
#include <windows.h>
#include <winsock2.h>
#include "ltrapitypes.h"
#include "ltrapidefine.h"
#include "critical.h"

//  
#define LC_VERMAJOR 4   //   
#define LC_VERMINOR 8   //   
#define LC_CURRENT_VERSION  ((LC_VERMAJOR << 16) | LC_VERMINOR)

#define LUSB_OK                                 0
#define LUSB_ERROR_INVALID_PARAMETERS           -1001
#define LUSB_ERROR_OPEN_DEVICE                  -1002
#define LUSB_ERROR_CLOSE_DEVICE                 -1003
#define LUSB_ERROR_DEVICE_CLOSED                -1004
#define LUSB_ERROR_CANCEL_IO                    -1005
#define LUSB_ERROR_TIMEOUT_EXPIRED              -1006
#define LUSB_ERROR_ASYNC_ERROR                  -1007
#define LUSB_ERROR_ASYNC_DATA_TRNSFER_LENGTH    -1008
#define LUSB_ERROR_ASYNC_DATA_TRNSFER           -1009

/*   LoadFPGA */
#define LTR_FPGA_ERR_BADPARAM   1
#define LTR_FPGA_ERR_FILEIO     2
#define LTR_FPGA_ERR_INITDMAC   10
#define LTR_FPGA_ERR_INITDMAC0  11
#define LTR_FPGA_ERR_INITDMAC7  12
#define LTR_FPGA_ERR_RESETFPGA  20
#define LTR_FPGA_ERR_INITFPGA   21
#define LTR_FPGA_ERR_PUTARRAY   22
#define LTR_FPGA_ERR_READBULK   30
#define LTR_FPGA_ERR_BADREPLY   31
#define LTR_FPGA_ERR_NOREPLY    32

/*    Ethernet */
#define LTR_IP_CMD_INIT         0x80000000
#define LTR_IP_CMD_NOP          0x00000000
#define LTR_IP_CMD_GET_NAME     0x00000001
#define LTR_IP_CMD_GET_ARRAY    0x00000002
#define LTR_IP_CMD_PUT_ARRAY    0x00000003

/*   TCP () */
#define LTR_IPERR_OK            0x00000000
#define LTR_IPERR_INVALID_PARAM 0x10000000
#define LTR_IPERR_CONNECT       0x20000000
#define LTR_IPERR_CONN_LOST     0x21000000
#define LTR_IPERR_NOT_CONNECTED 0x22000000
#define LTR_IPERR_TIMEOUT_R     0x30000000
#define LTR_IPERR_TIMEOUT_W     0x31000000
#define LTR_IPERR_READ          0x40000000
#define LTR_IPERR_WRITE         0x41000000
#define LTR_IPERR_ABORT         0x50000000
#define LTR_IPERR_BAD_SIG       0x60000000
#define LTR_IPERR_BAD_REPLYLEN  0x61000000
#define LTR_IPERR_BAD_REPLYDATA 0x62000000
#define LTR_IPERR_CMD_FAILED    0x70000000
#define LTR_IPERR_INTERNAL      0x7F000000

#define LTR_IP_DATA_PORT        11110
#define LTR_IP_IDENTIFY_PORT    11112
#define LTR_IP_CTRL_PORT        11113

#define LTR_IP_RESULT_OK        0x00000000
#define LTR_IP_RESULT_SYSFAIL   0xFFFFFFFF

#define LTR_IP_DEFAULT_IO_TIMEOUT    2000
#define LTR_IP_DEFAULT_CONN_TIMEOUT  60000

extern "C" DWORD WINAPI LTRAPI_LL_GetDLLVersion(void);

//-----------------------------------------------------------------------------
//------------------------------- LTR010 --------------------------------------
//-----------------------------------------------------------------------------
#pragma pack(4)
typedef struct
    {
    TDESCRIPTION_MODULE     Module;
    TDESCRIPTION_CPU        Cpu;
    TDESCRIPTION_FPGA       Fpga;
    TDESCRIPTION_INTERFACE  Interface;
    }
    TDESCRIPTION_LTR010;
#pragma pack()
//-----------------------------------------------------------------------------
class TLTR010_BOOT_LOADER
{
 public :
    HANDLE MainDevice;
    bool WINAPI SetUSBDevice(HANDLE Device);
    // /
    TLTR010_BOOT_LOADER();
    virtual ~TLTR010_BOOT_LOADER();

    virtual bool WINAPI CallApplication(DWORD address);

    virtual bool WINAPI GetArray(DWORD address, BYTE *data, DWORD size);
    virtual bool WINAPI PutArray(DWORD address, BYTE *data, DWORD size);

    //   
	virtual bool WINAPI GetModuleDescription(TDESCRIPTION_LTR010 *description);
	virtual bool WINAPI SetModuleDescription(TDESCRIPTION_LTR010 *description);
};
//-----------------------------------------------------------------------------
class TLTR010
{
 protected :
    BYTE fpga_info[FPGA_INFO_SIZE];
    HANDLE MainDevice;

 public :
    // /
    TLTR010();
    virtual ~TLTR010();

    bool NoRealLoadFPGAGlag;

    virtual bool WINAPI SetUSBDevice(HANDLE Device);

    virtual bool WINAPI GetArray(DWORD address, BYTE *data, DWORD size);
    virtual bool WINAPI PutArray(DWORD address, BYTE *data, DWORD size);
    //   
    virtual bool WINAPI GetModuleDescription(TDESCRIPTION_LTR010 *description);
    //     
    virtual bool WINAPI ResetFPGA(void);
    virtual bool WINAPI InitFPGA(void);
    virtual bool WINAPI InitDMAC(BYTE RDMA_PARAM, BYTE WDMA_PARAM);
    virtual bool WINAPI TLTR010::SetWaitTime(WORD mks100);
    virtual bool WINAPI LoadFPGA(BYTE *data, BYTE RDMA_PARAM, BYTE WDMA_PARAM, LPINT error_code);
    virtual bool WINAPI LoadFPGAFromFile(PCHAR fname, BYTE RDMA_PARAM, BYTE WDMA_PARAM, LPINT error_code);
    virtual bool WINAPI ConvertFPGADescription(TDESCRIPTION_LTR010 *description);
};

class TLTRUSB
{
    private:
        HANDLE hDevice;
        HANDLE MainDeviceClass;
    public:
        bool application_active;
        TLTRUSB();
        ~TLTRUSB();
        TLTR010 * Application;
        TLTR010_BOOT_LOADER * Bootloader;

        virtual INT WINAPI GetLastIoctlRetries(void);
        virtual HANDLE WINAPI GetUSBDeviceHandle(void);
        virtual void WINAPI CancelIO(void);
        virtual bool WINAPI GetModuleName(char mname[], DWORD Length);
        virtual bool WINAPI OpenLDeviceByName(LPTSTR DeviceName);
        virtual bool WINAPI CloseLDevice(void);
        virtual bool WINAPI SwitchBootloader(bool Bootloader);
};

extern "C" TLTRUSB * WINAPI CreateInstanceUSB(void);
extern "C" void WINAPI DeleteInstanceUSB(TLTRUSB *instance);

//-----------------------------------------------------------------------------
//------------------------------- LTR030 --------------------------------------
//-----------------------------------------------------------------------------
class TLTRIP
{
    private:
        TCritical cs;
        DWORD TimeoutConn;
        DWORD TimeoutCmd;
        DWORD LastErrorCode;
        DWORD ip_addr;
        BYTE fpga_info[FPGA_INFO_SIZE];
        WSAEVENT hAbortEvent;
        HANDLE hIdleTimer;
        LARGE_INTEGER IdleInterval;
        SOCKET ctlsock;
        WSAEVENT sock_evt;
        bool Connected;
        WORD protocol_version;

        bool __fastcall ExecEnd(DWORD StatusCode, DWORD SysError = 0);

    public:
        TLTRIP(WSAEVENT AbortEvent, DWORD IdleTimeout);
        ~TLTRIP();

        virtual void WINAPI SetTimeouts(DWORD ConnectTimeout, DWORD CommandTimeout);
        virtual bool WINAPI OpenDeviceByIP(DWORD ip);
        virtual bool WINAPI CloseDevice(void);
        virtual int WINAPI ConnectDataSocket(SOCKET sock);
        virtual int WINAPI GetLastErrorCode(void);

        virtual bool WINAPI Ping(DWORD Timeout = 0);
        virtual bool WINAPI GetModuleName(char * ModuleName, int ModuleNameLength);
        virtual bool WINAPI GetModuleDescription(TDESCRIPTION_LTR010 *description);
        virtual bool WINAPI GetArray(DWORD address, BYTE *data, DWORD size);
        virtual bool WINAPI PutArray(DWORD address, BYTE *data, DWORD size);

        virtual void WINAPI RestartIdleTimer(void);
        virtual void WINAPI StopIdleTimer(void);
        virtual HANDLE WINAPI GetIdleTimerHandle(void);

        virtual WORD WINAPI GetProtocolVersion(void);

        virtual bool WINAPI ExecCommand(DWORD CommandCode, DWORD CommandParam,
                                const BYTE *pSendBuf, DWORD SendSize,
                                BYTE *pRecvBuf, DWORD RecvSize,
                                DWORD *pResultCode,
                                DWORD *pReceivedBytes,
                                DWORD ForceTimeout = 0);

};

extern "C" TLTRIP * WINAPI CreateInstanceIP(WSAEVENT AbortEvent, DWORD IdleTimeout);
extern "C" void WINAPI DeleteInstanceIP(TLTRIP *instance);
#endif
