#include <vcl.h>
#include "LtrCrate.h"
#include "LtrClient.h"
#include "LtrServiceCrates.h"

#define PERMSTR(flag) ((flag) ? " permanently" : "")

/*------------------------------------------------------------------------------------------------*/
TClient::TClient(SOCKET _socket, DWORD peer_ip, TLTRServiceCrate* pCrateService)
    :
    socket(_socket),
    pcrate(NULL),
    pmodule(NULL),
    crate_channel(CC_CONTROL),
    control_thread(NULL),
    recv_buf_thread(NULL),
    send_buf_thread(NULL),
    flags(0),
    crate_service(pCrateService)
    {
    read_rec = TCrateModule::recv_zero;
    hTermEvent = CreateEvent(NULL, EVT_MANUAL_RESET, EVT_NONSIGNALED, NULL);
#if PER_CLIENT_CRATE_EVENTS
    hModuleGetEvent = CreateEvent(NULL, EVT_AUTO_RESET, EVT_NONSIGNALED, NULL);
    hModulePutEvent = CreateEvent(NULL, EVT_AUTO_RESET, EVT_NONSIGNALED, NULL);
#endif
    STRNCPYZ(peer_ip_str, IPToStr(peer_ip).c_str());

    //     (     )
    snprintf(clname, sizeof(clname), "%s [init]", peer_ip_str);

    control_thread = new TControlClientThread(this, false);
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
TClient::~TClient(void)
    {
    //   ,      
    SetEvent(hTermEvent);
    //    
    StopThread(control_thread);
    StopThread(send_buf_thread);
    StopThread(recv_buf_thread);
    Close();
#if PER_CLIENT_CRATE_EVENTS
    CloseHandle(hModulePutEvent);
    CloseHandle(hModuleGetEvent);
#endif
    CloseHandle(hTermEvent);
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
void TClient::Close(void)
    {
#if PER_CLIENT_CRATE_EVENTS
    if (pmodule)
        pmodule->DisconnectClient(this);
#endif
    SetEvent(hTermEvent);
    if (socket != INVALID_SOCKET)
        {
        shutdown(socket, SD_BOTH);
        closesocket(socket);
        socket = INVALID_SOCKET;
        }
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
__fastcall TControlClientThread::TControlClientThread(TClient *_client, bool CreateSuspended)
    : TThread(true), client(_client)
    {
    Priority = tpNormal;
    FreeOnTerminate = false;
    hEventRead = WSACreateEvent();
    hEventWrite = WSACreateEvent();
    hRawTimer = CreateWaitableTimer(NULL, TRUE, NULL);
    if (!CreateSuspended) Resume();
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
__fastcall TControlClientThread::~TControlClientThread()
    {
    CloseHandle(hRawTimer);
    CloseHandle(hEventWrite);
    CloseHandle(hEventRead);
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
int __fastcall TControlClientThread::recv(BYTE *buf, DWORD size)
    {
    if (!size || !buf) return 0;
    WSABUF wb;
    WSAOVERLAPPED ov;
    DWORD rcv_bytes = 0;
    bool ok;
    const char* clname = client->clname;

    ZeroMemory(&ov, sizeof(ov));
    ov.hEvent = hEventRead;

    while ((rcv_bytes < size) && !Terminated && !client->IsTerminating())
        {
        wb.len = size - rcv_bytes;
        wb.buf = buf + rcv_bytes;
        DWORD xfer_size = 0;
        DWORD flag = 0;
        int res = WSARecv(client->socket, &wb, 1, &xfer_size, &flag, &ov, NULL);
        ok = (res != SOCKET_ERROR);
        if (!ok)
            {
            int err = WSAGetLastError();
            if (err == WSA_IO_PENDING)
                {
                HANDLE evt[2] = { hEventRead, client->hTermEvent };
                switch (WSAWaitForMultipleEvents(2, evt, FALSE, INFINITE, FALSE))
                    {
                    case WSA_WAIT_EVENT_0:  //  
                        ok = WSAGetOverlappedResult(client->socket, &ov, &xfer_size, FALSE, &flag);
                        if (!ok)
                            { //    
                            err = WSAGetLastError();
                            LOGWRITE(Format("%s: async receive failed (error %d)",
                                ARRAYOFCONST((clname, err))),
                                LOGTASK_CLIENT_CTL, LOGL_DEBUG1);
                            client->RequestTermination();
                            }
                        break;
                    case WSA_WAIT_EVENT_0 + 1:  //  
                        CancelIo((HANDLE)client->socket);
                        WSAGetOverlappedResult(client->socket, &ov, &xfer_size, TRUE, &flag);
                        break;
                    default:                // 
                        CancelIo((HANDLE)client->socket);
                        WSAGetOverlappedResult(client->socket, &ov, &xfer_size, TRUE, &flag);
                        LOGWRITE(Format("%s: receive failed (wait error)",
                            ARRAYOFCONST((clname))),
                            LOGTASK_CLIENT_CTL, LOGL_DEBUG1);
                        client->RequestTermination();
                        break;
                    } /* switch */
                }
            else
                { //   , 
                LOGWRITE(Format("%s: receive failed (error %d)",
                    ARRAYOFCONST((clname, err))),
                    LOGTASK_CLIENT_CTL, LOGL_DEBUG1);
                client->RequestTermination();
                }
            }

        if (!ok) break;
        // :   , xfer_size =  
        if (xfer_size == 0)
            {
            LOGWRITE(Format("%s: received end-of-connection",
                ARRAYOFCONST((clname))),
                LOGTASK_CLIENT_CTL, LOGL_DEBUG1);
            client->RequestTermination();
            break;
            }
        rcv_bytes += xfer_size;
        } /* while */
    return rcv_bytes;
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
int __fastcall TControlClientThread::send(BYTE *buf, DWORD size)
    {
    if (!size || !buf) return 0;
    WSABUF wb;
    WSAOVERLAPPED ov;
    bool ok;
    const char* clname = client->clname;

    ZeroMemory(&ov, sizeof(ov));
    ov.hEvent = hEventWrite;

    //   winsock,  WSASend()     .
    //    .
    // ,    ,    .

    if (Terminated || client->IsTerminating())
        return 0;

    wb.len = size;
    wb.buf = buf;
    DWORD xfer_size = 0;
    DWORD flag = 0;
    int res = WSASend(client->socket, &wb, 1, &xfer_size, 0, &ov, NULL);
    ok = (res != SOCKET_ERROR);
    if (!ok)
        {
        int err = WSAGetLastError();
        if (err == WSA_IO_PENDING)
            {
            HANDLE evt[2] = { hEventWrite, client->hTermEvent };
            switch (WSAWaitForMultipleEvents(2, evt, FALSE, INFINITE, FALSE))
                {
                case WSA_WAIT_EVENT_0:  //  
                    ok = WSAGetOverlappedResult(client->socket, &ov, &xfer_size, FALSE, &flag);
                    if (!ok)
                        { //    
                        err = WSAGetLastError();
                        LOGWRITE(Format("%s: async send failed (error %d)",
                            ARRAYOFCONST((clname, err))),
                            LOGTASK_CLIENT_CTL, LOGL_DEBUG1);
                        client->RequestTermination();
                        }
                    break;
                case WSA_WAIT_EVENT_0 + 1:  //  
                    CancelIo((HANDLE)client->socket);
                    WSAGetOverlappedResult(client->socket, &ov, &xfer_size, TRUE, &flag);
                    break;
                default:                // 
                    CancelIo((HANDLE)client->socket);
                    WSAGetOverlappedResult(client->socket, &ov, &xfer_size, TRUE, &flag);
                    LOGWRITE(Format("%s: send failed (wait error)",
                        ARRAYOFCONST((clname))),
                        LOGTASK_CLIENT_CTL, LOGL_DEBUG1);
                    client->RequestTermination();
                    break;
                } /* switch */
            }
        else
            { //   , 
            LOGWRITE(Format("%s: send failed (error %d)",
                ARRAYOFCONST((clname, err))),
                LOGTASK_CLIENT_CTL, LOGL_DEBUG1);
            client->RequestTermination();
            }
        }

    return (ok) ? xfer_size : 0;
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
void __fastcall TControlClientThread::ack(bool ack_flag)
    {
    DWORD ack_value = ack_flag ? CONTROL_ACKNOWLEDGE_GOOD : CONTROL_ACKNOWLEDGE_BAD;
    send((BYTE*)&ack_value, CONTROL_ACKNOWLEDGE_SIZE);
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
bool __fastcall TControlClientThread::ProcessStartup()
    {
    AnsiString connrefused = 
        Format("Connection refused for client %s: ", ARRAYOFCONST((client->peer_ip_str)));

    bool ack_flag = false;
    // request[] = cmd[8] + csn[16] + cc[2] + tmark[4]
    BYTE request[CONTROL_COMMAND_SIZE+16+2+4];

    /*       terminate */
    if ((sizeof(request) == recv(request, sizeof(request))) &&
        *(DWORD*)(request+0)==CONTROL_COMMAND_START &&
        *(DWORD*)(request+4)==CONTROL_COMMAND_INIT_CONNECTION)
        {
        WORD Client_flags = (*(WORD*)(request + CONTROL_COMMAND_SIZE + 16)) & 0xFF00;
        WORD Crate_Channel = (*(WORD*)(request + CONTROL_COMMAND_SIZE + 16)) & 0x00FF; //   
        char* Request_CSN = request + CONTROL_COMMAND_SIZE;

        if ((Crate_Channel == CC_CONTROL) || (Crate_Channel == 1 + MODULE_INDEX_USER)
            || ((Crate_Channel > 0) && (Crate_Channel <= MODULE_MAX)))
            { //   
            /*       API ,      */
            if (!strncmp(Request_CSN, CSN_SERVER_CONTROL, 16))
                { /* server control API */
                if (Crate_Channel == CC_CONTROL)
                    {
                    ack(true);
                    send(request, sizeof(request));
                    client->pcrate = NULL;
                    client->crate_channel = CC_CONTROL;
                    client->flags = Client_flags;
                    snprintf(client->clname, sizeof(client->clname), "%s [server control]",
                        client->peer_ip_str);
                    ack_flag = true;
                    }
                else LOGWRITE(connrefused + "server control requires cc == CC_CONTROL.",
                    LOGTASK_CLIENT_INIT, LOGL_ERROR);
                }
            else
                { /* not server control API */
                if (WaitForSingleObject(crate_list_access, INFINITE) == WAIT_OBJECT_0)
                    {
                    bool already_open = false;
                    TCrate* found_crate = NULL;
                    for (int crate_index = 0; crate_index < CRATE_MAX; crate_index++)
                        {
                        if ((crate_list[crate_index]->GetState() == CRATE_ST_ONLINE) &&
                            //         
                            ((!*Request_CSN) || !strncmp(Request_CSN, crate_list[crate_index]->csn, 16))
                            )
                            {
                            if (Crate_Channel == CC_CONTROL)
                                {
                                found_crate = crate_list[crate_index];
                                break;
                                }
                            if (crate_list[crate_index]->Module[Crate_Channel - 1].Present())
                                {
                                /*   (  ),      */
                                found_crate = crate_list[crate_index];
                                if (WaitForSingleObject(client_list_access, INFINITE) == WAIT_OBJECT_0)
                                    {
                                    for (int i = 0; i < CLIENT_MAX; i++)
                                        {
                                        if (client_list[i] && (client_list[i]->pcrate == found_crate)
                                           && (client_list[i]->crate_channel == Crate_Channel)
                                           && client_list[i]->IsOpen())
                                           {
                                           already_open = true;
                                           break;
                                           }
                                        }
                                    ReleaseMutex(client_list_access);
                                    }
                                break;
                                }
                            }
                        } /* for crate_index */

                    if  (
                        (found_crate)
                        &&
                        ((!(found_crate->sflags & CRATE_FLAG_BOOT_LOADER)) || (Crate_Channel == CC_CONTROL))
                        )
                        {
                        /*  .        ,  ,
                             */
                        if (!already_open || g_Conf.Multiconnect())
                            {
                            //    
                            memcpy(Request_CSN, found_crate->csn, 16);
                            //    
                            DWORD tmark = ((found_crate->start_mark_counter & 0xFFFF) << 16)|
                                (found_crate->second_mark_counter & 0xFFFF);
                            *(DWORD*)(request + CONTROL_COMMAND_SIZE + 16 + 2) = tmark;

                            //         / 
                            if (!already_open)
                                {
                                ack(true);
                                }
                            else
                                {//   
                                DWORD ack_value = CONTROL_ACKNOWLEDGE_MODULE_IN_USE;
                                send((BYTE*)&ack_value, CONTROL_ACKNOWLEDGE_SIZE);
                                LOGWRITE(Format("Warning! Client %s [%.16s : %d]: module already open"
                                    " (connection accepted, set multiconnect=0 to deny such attempts).",
                                    ARRAYOFCONST((client->peer_ip_str, Request_CSN, Crate_Channel))),
                                    LOGTASK_CLIENT_INIT, LOGL_WARNING);
                                }
                            send(request, sizeof(request));
                            client->pcrate = found_crate;
                            client->crate_channel = Crate_Channel;
                            client->flags = Client_flags;
                            snprintf(client->clname, sizeof(client->clname), "%s [%s : %d]",
                                client->peer_ip_str, found_crate->csn, Crate_Channel);
                            ack_flag = true;
                            }
                        else
                            {
                            LOGWRITE(Format(connrefused + "module [%.16s : %d] already open.",
                                ARRAYOFCONST((Request_CSN, Crate_Channel))),
                                LOGTASK_CLIENT_INIT, LOGL_ERROR);
                            }
                        }
                    else
                        {
                        LOGWRITE(Format(connrefused + "module [%.16s : %d] not found.",
                            ARRAYOFCONST((Request_CSN, Crate_Channel))),
                            LOGTASK_CLIENT_INIT, LOGL_ERROR);
                        }
                    ReleaseMutex(crate_list_access);
                    }
                else LOGWRITE(connrefused + "cannot access crate list.",
                    LOGTASK_CLIENT_INIT, LOGL_ERROR);
                } /* if server control */
            } /* channel */
        else LOGWRITE(connrefused + "invalid channel number " + IntToStr(Crate_Channel) + ".",
            LOGTASK_CLIENT_INIT, LOGL_ERROR);
        } /* request valid */
    else
        {
        LOGWRITE(connrefused + "error in request.", LOGTASK_CLIENT_INIT, LOGL_ERROR);
        //      
        ack_flag = true;
        }

    if (!ack_flag)
        ack(false);
    return ack_flag;
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
void __fastcall TControlClientThread::Execute()
    {
    /*      Control Thread,     .
       ,     ,   - ,  
         ,     */

    /* -------- STARTUP -------- */
    if (!ProcessStartup())
        {
        client->Close();
        return;
        }

    TCrate* pcrate = client->pcrate;
    const char* clname = client->clname;

    LOGWRITE(Format("%s: Connection started", ARRAYOFCONST((clname))), LOGTASK_CLIENT_CTL, LOGL_INFO);

    if (client->crate_channel != CC_CONTROL)
        {
        //    control,  data
        client->pmodule = &pcrate->Module[client->crate_channel - 1];
#if PER_CLIENT_CRATE_EVENTS
        if (client->pmodule->ConnectClient(client) != TCrateModule::eOk)
#else
        if (client->pmodule->ClientGetInit(client->read_rec) != TCrateModule::eOk)
#endif
            {
            client->pmodule = NULL;
            }
        client->recv_buf_thread = new TReadClientThread(client, true);
        client->send_buf_thread = new TWriteClientThread(client, true);
        client->send_buf_thread->Resume();
        client->recv_buf_thread->Resume();
        /*   ,  control  */
        return;
        }

    if (pcrate && (client->flags & CC_RAW_DATA_FLAG))
        {
        client->pmodule = &pcrate->Module[MODULE_INDEX_RAW];
#if PER_CLIENT_CRATE_EVENTS
        if (client->pmodule->ConnectClient(client) == TCrateModule::eOk)
#else
        if (client->pmodule->ClientGetInit(client->read_rec) == TCrateModule::eOk)
#endif
            {
            pcrate->ReadRawData = true;
            }
        else
            {
            client->pmodule = NULL;
            }
        }

    /* -------- CONTROL THREAD -------- */

    const DWORD RawBufLen = CLIENT_READ_BUF_LEN;
    DWORD* pRawDataBuffer = new DWORD[RawBufLen];

    while (!Terminated && !client->IsTerminating())
        {
        bool ack_flag = false;
        int res = recv((BYTE*)cheader, 8);
        if (res == 0) break;

        if ((res == 8)
            && (cheader[0] == CONTROL_COMMAND_START)
            && ((cheader[1] & CONTROL_COMMAND_BASE_MASK_) == CONTROL_COMMAND_BASE_))
            {
            DWORD CommandCode = cheader[1];
            switch (CommandCode)
                {
                case CONTROL_COMMAND_BOOT_LOADER_PUT_ARRAY :
                case CONTROL_COMMAND_PUT_ARRAY :
                    // DWORD addr, size
                    res = recv((BYTE *)cheader, 8);
                    if ((res == 8) && (cheader[1] <= sizeof(ControlBuffer)))
                        {
                        res = recv(ControlBuffer, cheader[1]);
                        if (res == (int)cheader[1])
                            {
                            if (pcrate)
                                {
                                LOGWRITE(Format("%s: PutArray addr = %0.8X, size = %d",
                                    ARRAYOFCONST((clname, cheader[0], cheader[1]))),
                                    LOGTASK_CLIENT_CTL, LOGL_DEBUG1);
                                ack_flag = pcrate->PutArray(
                                    (CommandCode == CONTROL_COMMAND_PUT_ARRAY) ? CM_APP : CM_BOOT,
                                    cheader[0], ControlBuffer, cheader[1]);
                                }
                            else
                                {
                                LOGWRITE(Format("%s: PutArray not supported for control API",
                                    ARRAYOFCONST((clname))),
                                    LOGTASK_CLIENT_CTL, LOGL_WARNING);
                                ack_flag = false;
                                }
                            ack(ack_flag);
                            }
                        }
                    break;

                case CONTROL_COMMAND_BOOT_LOADER_GET_ARRAY :
                case CONTROL_COMMAND_GET_ARRAY :
                    // DWORD addr, size
                    res = recv((BYTE *)cheader, 8);
                    if ((res == 8) && (cheader[1] <= sizeof(ControlBuffer)))
                        {
                        if (pcrate)
                            {
                            LOGWRITE(Format("%s: GetArray addr = %0.8X, size = %d",
                                ARRAYOFCONST((clname, cheader[0], cheader[1]))),
                                LOGTASK_CLIENT_CTL, LOGL_DEBUG1);
                            ack_flag = pcrate->GetArray(
                                (CommandCode == CONTROL_COMMAND_GET_ARRAY) ? CM_APP : CM_BOOT,
                                cheader[0], ControlBuffer, cheader[1]);
                            }
                        else
                            {
                            LOGWRITE(Format("%s: GetArray not supported for control API",
                                ARRAYOFCONST((clname))),
                                LOGTASK_CLIENT_CTL, LOGL_WARNING);
                            ack_flag = false;
                            }
                        ack(ack_flag);
                        if (ack_flag) send(ControlBuffer, cheader[1]);
                        }
                    break;

                case CONTROL_COMMAND_BOOT_LOADER_GET_DESCRIPTION :
                case CONTROL_COMMAND_GET_DESCRIPTION :
                    if (pcrate)
                        {
                        LOGWRITE(Format("%s: GetDescription", ARRAYOFCONST((clname))),
                            LOGTASK_CLIENT_CTL, LOGL_DEBUG1);
                        ack_flag = pcrate->GetModuleDescription(
                            (CommandCode == CONTROL_COMMAND_GET_DESCRIPTION) ? CM_APP : CM_BOOT,
                            (TDESCRIPTION_LTR010*)ControlBuffer);
                        }
                    else
                        {
                        LOGWRITE(Format("%s: GetDescription not supported for control API",
                            ARRAYOFCONST((clname))),
                            LOGTASK_CLIENT_CTL, LOGL_WARNING);
                        ack_flag = false;
                        }
                    ack(ack_flag);
                    if (ack_flag) send(ControlBuffer, sizeof(TDESCRIPTION_LTR010));
                    break;

                case CONTROL_COMMAND_BOOT_LOADER_SET_DESCRIPTION :
                    res = recv(ControlBuffer, sizeof(TDESCRIPTION_LTR010));
                    if (res == sizeof(TDESCRIPTION_LTR010))
                        {
                        if (pcrate)
                            {
                            LOGWRITE(Format("%s: SetDescription", ARRAYOFCONST((clname))),
                                LOGTASK_CLIENT_CTL, LOGL_DEBUG1);
                            ack_flag = pcrate->SetModuleDescription(CM_AUTO, (TDESCRIPTION_LTR010*)ControlBuffer);
                            }
                        else
                            {
                            LOGWRITE(Format("%s: SetDescription not supported for control API",
                                ARRAYOFCONST((clname))),
                                LOGTASK_CLIENT_CTL, LOGL_WARNING);
                            ack_flag = false;
                            }
                        ack(ack_flag);
                        }
                    break;

                case CONTROL_COMMAND_LOAD_FPGA :
                    // BYTE ControlBuffer[EP1K50_SIZE]
                    res = recv(ControlBuffer, EP1K50_SIZE + 2);
                    if (res == EP1K50_SIZE + 2)
                        {
                        if (pcrate)
                            {
                            //      
                            if (WaitForSingleObject(crate_list_access, INFINITE)==WAIT_OBJECT_0)
                                {
                                LOGWRITE(Format("%s: Reloading crate", ARRAYOFCONST((clname))),
                                    LOGTASK_CLIENT_CTL, LOGL_INFO);
                                ack_flag = pcrate->Startup(ControlBuffer + 2, false);

                                if (WaitForSingleObject(client_list_access, INFINITE)==WAIT_OBJECT_0)
                                    {
                                    if (!ack_flag)
                                        {
                                        LOGWRITE(Format("%s: Cannot reload crate", ARRAYOFCONST((clname))),
                                            LOGTASK_CRATE_INIT, LOGL_ERROR);
                                        client->RequestTermination();
                                        }
                                    ReleaseMutex(client_list_access);
                                    }
                                else
                                    {
                                    LOGWRITE(Format("%s: Cannot access client list", ARRAYOFCONST((clname))),
                                        LOGTASK_CLIENT_CTL, LOGL_ERROR);
                                    }
                                ReleaseMutex(crate_list_access);
                                }
                            else
                                {
                                LOGWRITE(Format("%s: Cannot access crate list", ARRAYOFCONST((clname))),
                                    LOGTASK_CLIENT_CTL, LOGL_ERROR);
                                }
                            }
                        else /* !pcrate */
                            {
                            LOGWRITE(Format("%s: ReloadCrate not supported for control API", ARRAYOFCONST((clname))),
                                LOGTASK_CLIENT_CTL, LOGL_WARNING);
                            ack_flag = false;
                            }

                        ack(ack_flag);
                        }
                    break;

                case CONTROL_COMMAND_CALL_APPLICATION:
                    res = recv((BYTE *)cheader, 4);
                    if (res == 4)
                        {
                        if (pcrate)
                            {
                            ack_flag = pcrate->CallApplication(cheader[0]);
                            }
                        else
                            {
                            LOGWRITE(Format("%s: CallApplication not supported for control API",
                                ARRAYOFCONST((clname))),
                                LOGTASK_CLIENT_CTL, LOGL_WARNING);
                            ack_flag = false;
                            }
                        ack(ack_flag);
                        }
                    break;

                case CONTROL_COMMAND_GET_CRATES:
                    memset(ControlBuffer, 0, CRATE_MAX*16);
                    if (WaitForSingleObject(crate_list_access, INFINITE)==WAIT_OBJECT_0)
                        {
                        for (int i = 0, j = 0; i < CRATE_MAX; i++)
                            {
                            if (crate_list[i]->GetState() == CRATE_ST_ONLINE)
                                {
                                if (crate_list[i]->sflags & CRATE_FLAG_BOOT_LOADER)
                                    {
                                    memset(ControlBuffer + j * NAME_SIZE, 'X', NAME_SIZE);
                                    }
                                else
                                    {
                                    strncpy(ControlBuffer + j * NAME_SIZE,
                                        crate_list[i]->csn, NAME_SIZE);
                                    }
                                j++;
                                }
                            }
                        ReleaseMutex(crate_list_access);
                        ack_flag=true;
                        }
                    else ack_flag=false;
                    ack(ack_flag);
                    if (ack_flag) send(ControlBuffer, CRATE_MAX * 16);
                    break;

                case CONTROL_COMMAND_GET_MODULES :
                    if (pcrate)
                        {
                        for (int i = 0; i < MODULE_MAX; i++)
                            {
                            pcrate->Module[i].GetMID(((WORD*)ControlBuffer)[i]);
                            }
                        ack_flag = true;
                        }
                    else
                        {
                        LOGWRITE(Format("%s: GetModules not supported for control API",
                            ARRAYOFCONST((clname))),
                            LOGTASK_CLIENT_CTL, LOGL_WARNING);
                        ack_flag = false;
                        }
                    ack(ack_flag);
                    if (ack_flag) send(ControlBuffer, MODULE_MAX * sizeof(WORD));
                    break;

                case CONTROL_COMMAND_GET_CRATE_RAW_DATA_SIZE:
                    if (pcrate && client->pmodule)
                        {
                        DWORD SendValue = 0;
                        client->read_rec.buf_len = TCrateModule::RecvBufLen;
                        client->pmodule->ClientGetLen(client->read_rec, SendValue);
                        send((BYTE*)&SendValue, sizeof(SendValue));
                        }
                    break;

                case CONTROL_COMMAND_GET_CRATE_RAW_DATA:
                    {
                    DWORD rbuf[3];
                    res = recv((BYTE *)rbuf, sizeof(rbuf));
                    if ((res == sizeof(rbuf)) && pcrate && client->pmodule)
                        {
                        bool error = false;
                        DWORD WordsLeft = rbuf[0];
                        DWORD Timeout = rbuf[1];
                        bool FullRawData = rbuf[2];
                        DWORD tmark = ((pcrate->start_mark_counter & 0xFFFF) << 16) |
                                       (pcrate->second_mark_counter & 0xFFFF);
                        DWORD OutOfs = 0;

                        if (tmark && pRawDataBuffer)
                            {
                            pRawDataBuffer[OutOfs++] = LTR_SIGNAL_TSTAMP;
                            pRawDataBuffer[OutOfs++] = tmark;
                            }

                        //  
                        LARGE_INTEGER Interval;
                        Interval.QuadPart = -10000LL * Timeout;
                        SetWaitableTimer(hRawTimer, &Interval, 0, NULL, NULL, FALSE);

                        //        
                        while ((WordsLeft > 0) & !error)
                            {
                            if (WaitForSingleObject(hRawTimer, 0) == WAIT_OBJECT_0)
                                break;
                            client->read_rec.buf = pRawDataBuffer + OutOfs;
                            client->read_rec.buf_len = RawBufLen - OutOfs;
                            DWORD WordsReady;
                            switch (client->pmodule->ClientGet(client->read_rec, WordsReady))
                                {
                                case TCrateModule::eOk:
                                    //  WordsReady 
                                    break;
                                case TCrateModule::eOverflow:
                                    //     
                                    WordsReady = std::min(WordsLeft, TCrateModule::RecvBufLen);
                                    client->read_rec.last_count -= WordsReady;
                                    client->read_rec.last_offset =
                                        (client->read_rec.last_offset - WordsReady)
                                        % TCrateModule::RecvBufLen;
                                    if (!FullRawData && (OutOfs < RawBufLen))
                                        pRawDataBuffer[OutOfs++] = LTR_SIGNAL_RBUF_OVF;
                                    //     
                                    WordsReady = 0;
                                    break;
                                default:
                                    //  --  
                                    WordsReady = 0;
                                    error = true;
                                    break;
                                }
                            DWORD StartOfs = OutOfs;
                            for (DWORD i = 0; i < WordsReady; i++)
                                {
                                DWORD Data = pRawDataBuffer[StartOfs + i];
                                if (((Data & LTR_SIG_ESC_MASK) != LTR_SIG_ESC_VAL)
                                    && (Data & LTR010CMD) && !FullRawData)
                                    { //  
                                    continue;
                                    }
                                pRawDataBuffer[OutOfs++] = Data;
                                }
                            // OutOfs =    
                            if (OutOfs > WordsLeft)
                                OutOfs = WordsLeft;
                            if (OutOfs > 0)
                                { //   
                                send((BYTE*)pRawDataBuffer, OutOfs * sizeof(DWORD));
                                }
                            else if (!error)
                                { //  ,    
                                HANDLE evt[2] =
                                    {
#if !PER_CLIENT_CRATE_EVENTS
                                    client->pmodule->hCratePutEvent,
#else
                                    client->hModulePutEvent,
#endif
                                    hRawTimer
                                    };
                                if (WaitForMultipleObjects(2, evt, FALSE, Timeout) != WAIT_OBJECT_0)
                                    break;
                                }
                            WordsLeft -= OutOfs;
                            OutOfs = 0;
                            } /* while */
                        } /* if res == ... */
                    else ack(false);
                    }
                    break;

                case CONTROL_COMMAND_GET_CRATE_TYPE:
                    if (pcrate)
                        {
                        ack(true);
                        TCRATE_INFO CrateInfo;
                        CrateInfo.CrateType = pcrate->CrateType;
                        CrateInfo.CrateInterface = pcrate->GetConnType();
                        send((BYTE *)&CrateInfo, sizeof(CrateInfo));
                        }
                    else
                        {
                        ack(false);
                        LOGWRITE(Format("%s: GetCrateType not supported for control API",
                            ARRAYOFCONST((clname))),
                            LOGTASK_CLIENT_CTL, LOGL_WARNING);
                        }
                    break;

                //    IP-,    
                case CONTROL_COMMAND_IP_GET_LIST:
                    {
                    const int ListType[] = { IPLIST_STATIC, IPLIST_DYNAMIC };
                    struct //    (  )
                        {
                        DWORD max_entries; //     
                        DWORD netaddr;     //  IP  (host-endian)
                        DWORD netmask;     //  IP  (host-endian)
                        }
                        params;
                    struct //    ()
                        {
                        DWORD num_found;    //    
                        DWORD num_returned; //   = min(max_entries, num_found)
                        //    num_returned  TIPCRATE_ENTRY
                        }
                        reply_header;
                    std::vector<TIPCRATE_ENTRY> out_vec;

                    res = recv((BYTE*)&params, sizeof(params));
                    if (res == sizeof(params))
                        {
                        reply_header.num_found = reply_header.num_returned = 0;
                        for (int type = 0; type < sizeof(ListType) / sizeof(ListType[0]); type++)
                            {
                            TAddrList *IPList =
                                IPCrateList.ExportList(ListType[type]);
                            if (!IPList) continue;
                            for (TAddrListNode* addr = IPList->GetFirstNode(); addr; addr = addr->next)
                                {
                                TIPCRATE_ENTRY entry;
                                entry.ip_addr = ntohl(addr->dwIP);
                                //     
                                if ((entry.ip_addr & params.netmask) != (params.netaddr & params.netmask))
                                    continue;
                                //  
                                //   ,    
                                if (++reply_header.num_found > params.max_entries)
                                    continue;
                                entry.is_dynamic = (ListType[type] == IPLIST_DYNAMIC);
                                entry.flags = addr->dwFlags;
                                strncpy(entry.serial_number, addr->szSerial, sizeof(entry.serial_number));
                                switch (addr->iStatus)
                                    {
                                    case IP_STATUS_OFFLINE:
                                        entry.status = CRATE_IP_STATUS_OFFLINE;
                                        break;
                                    case IP_STATUS_CONNECTING:
                                        entry.status = CRATE_IP_STATUS_CONNECTING;
                                        break;
                                    case IP_STATUS_ONLINE:
                                        entry.status = CRATE_IP_STATUS_ONLINE;
                                        break;
                                    default:
                                        entry.status = CRATE_IP_STATUS_ERROR;
                                        break;
                                    }
                                out_vec.push_back(entry);
                                }
                            delete IPList;
                            } /* for type */

                        ack_flag = true;
                        }
                    else
                        {
                        LOGWRITE(Format("%s: Get IP address list: bad request",
                            ARRAYOFCONST((clname))),
                            LOGTASK_CLIENT_CTL, LOGL_ERROR);
                        }
                    ack(ack_flag);
                    if (ack_flag)
                        {
                        reply_header.num_returned = out_vec.size();
                        send((BYTE*)&reply_header, sizeof(reply_header));
                        if (reply_header.num_returned > 0)
                            send((BYTE*)&out_vec[0], sizeof(TIPCRATE_ENTRY) * reply_header.num_returned);
                        }
                    }
                    break;

                //  IP- 
                case CONTROL_COMMAND_IP_ADD_LIST_ENTRY:
                    {
                    struct //    (  )
                        {
                        DWORD ip_addr;     // IP- (host-endian)
                        DWORD flags;       //  
                        DWORD permanent;   //   ini 
                        }
                        params;
                    res = recv((BYTE*)&params, sizeof(params));
                    if (res == sizeof(params))
                        {
                        DWORD ip = htonl(params.ip_addr);
                        AnsiString ip_str = IPToStr(ip);
                        if (g_Conf.AddIPCrate(IPCrateList, ip,
                            params.flags & CRATE_IP_FLAG__VALID_BITS_,
                            params.permanent))
                            {
                            client->crate_service->PostDisplayMessage(LTR_DISP_ip_list_updated);
                            LOGWRITE(Format("%s: Add IP: %s added%s",
                                ARRAYOFCONST((clname, ip_str, PERMSTR(params.permanent)))),
                                LOGTASK_CLIENT_CTL, LOGL_VERBOSE);
                            ack_flag = true;
                            }
                        else
                            {
                            LOGWRITE(Format("%s: Add IP: cannot add %s",
                                ARRAYOFCONST((clname, ip_str))),
                                LOGTASK_CLIENT_CTL, LOGL_ERROR);
                            }
                        }
                    else
                        {
                        LOGWRITE(Format("%s: Add IP: bad request", ARRAYOFCONST((clname))),
                            LOGTASK_CLIENT_CTL, LOGL_ERROR);
                        }
                    ack(ack_flag);
                    }
                    break;

                //  IP- 
                case CONTROL_COMMAND_IP_DELETE_LIST_ENTRY:
                    {
                    struct //    (  )
                        {
                        DWORD ip_addr;     // IP- (host-endian)
                        DWORD permanent;   //   ini 
                        }
                        params;
                    res = recv((BYTE*)&params, sizeof(params));
                    if (res == sizeof(params))
                        {
                        TAddrListNode node;
                        DWORD ip = htonl(params.ip_addr);
                        AnsiString ip_str = IPToStr(ip);
                        switch (IPCrateList.Find(ip, &node))
                            {
                            case IPLIST_STATIC:
                                ack_flag = g_Conf.DeleteIPCrate(IPCrateList, ip, params.permanent);
                                break;
                            case IPLIST_DYNAMIC:
                                ack_flag = IPCrateList.Delete(IPLIST_DYNAMIC, ip);
                                break;
                            default:
                                ack_flag = false;
                            }
                        if (ack_flag)
                            {
                            client->crate_service->PostDisplayMessage(LTR_DISP_ip_list_updated);
                            LOGWRITE(Format("%s: Delete IP: %s deleted%s",
                                ARRAYOFCONST((clname, ip_str, PERMSTR(params.permanent)))),
                                LOGTASK_CLIENT_CTL, LOGL_VERBOSE);
                            }
                        else
                            {
                            LOGWRITE(Format("%s: Delete IP: cannot delete %s",
                                ARRAYOFCONST((clname, ip_str))),
                                LOGTASK_CLIENT_CTL, LOGL_ERROR);
                            }
                        }
                    else
                        {
                        LOGWRITE(Format("%s: Delete IP: bad request", ARRAYOFCONST((clname))),
                            LOGTASK_CLIENT_CTL, LOGL_ERROR);
                        }
                    ack(ack_flag);
                    }
                    break;

                //  IP 
                case CONTROL_COMMAND_IP_CONNECT:
                    {
                    struct //    (  )
                        {
                        DWORD ip_addr;     // IP- (host-endian)
                        }
                        params;
                    res = recv((BYTE*)&params, sizeof(params));
                    if (res == sizeof(params))
                        {
                        TAddrListNode node;
                        DWORD ip = htonl(params.ip_addr);
                        AnsiString ip_str = IPToStr(ip);
                        int list_idx = IPCrateList.Find(ip, &node);
                        if (list_idx >= 0)
                            {
                            switch (node.iStatus)
                                {
                                case IP_STATUS_OFFLINE:
                                    client->crate_service->PostControlMessage(
                                        MsgControl(LTR_SRV_connect_crate, node.dwIP));
                                    LOGWRITE(Format("%s: Connect IP: connecting %s",
                                        ARRAYOFCONST((clname, ip_str))),
                                        LOGTASK_CLIENT_CTL, LOGL_VERBOSE);
                                    ack_flag = true;
                                    break;
                                case IP_STATUS_CONNECTING:
                                case IP_STATUS_ONLINE:
                                    LOGWRITE(Format("%s: Connect IP: %s already connected or connecting",
                                        ARRAYOFCONST((clname, ip_str))),
                                        LOGTASK_CLIENT_CTL, LOGL_WARNING);
                                    ack_flag = true;
                                    break;
                                default:
                                    LOGWRITE(Format("%s: Connect IP: cannot connect %s",
                                        ARRAYOFCONST((clname, ip_str))),
                                        LOGTASK_CLIENT_CTL, LOGL_ERROR);
                                    break;
                                }
                            }
                        else
                            {
                            LOGWRITE(Format("%s: Connect IP: %s not found",
                                ARRAYOFCONST((clname, ip_str))),
                                LOGTASK_CLIENT_CTL, LOGL_ERROR);
                            }
                        }
                    else
                        {
                        LOGWRITE(Format("%s: Connect IP: bad request", ARRAYOFCONST((clname))),
                            LOGTASK_CLIENT_CTL, LOGL_ERROR);
                        }
                    ack(ack_flag);
                    }
                    break;

                //  IP 
                case CONTROL_COMMAND_IP_DISCONNECT:
                    {
                    struct //    (  )
                        {
                        DWORD ip_addr;     // IP- (host-endian)
                        }
                        params;
                    res = recv((BYTE*)&params, sizeof(params));
                    if (res == sizeof(params))
                        {
                        TAddrListNode node;
                        DWORD ip = htonl(params.ip_addr);
                        AnsiString ip_str = IPToStr(ip);
                        int list_idx = IPCrateList.Find(ip, &node);
                        if (list_idx >= 0)
                            {
                            switch (node.iStatus)
                                {
                                case IP_STATUS_OFFLINE:
                                    LOGWRITE(Format("%s: Disconnect IP: %s already disconnected",
                                        ARRAYOFCONST((clname, ip_str))),
                                        LOGTASK_CLIENT_CTL, LOGL_WARNING);
                                    ack_flag = true;
                                    break;
                                case IP_STATUS_CONNECTING:
                                case IP_STATUS_ONLINE:
                                    client->crate_service->PostControlMessage(
                                        MsgControl(LTR_SRV_disconnect_crate, node.dwIP));
                                    LOGWRITE(Format("%s: Disconnect IP: disconnecting %s",
                                        ARRAYOFCONST((clname, ip_str))),
                                        LOGTASK_CLIENT_CTL, LOGL_VERBOSE);
                                    ack_flag = true;
                                    break;
                                default:
                                    LOGWRITE(Format("%s: Disconnect IP: cannot disconnect %s",
                                        ARRAYOFCONST((clname, ip_str))),
                                        LOGTASK_CLIENT_CTL, LOGL_ERROR);
                                    break;
                                }
                            }
                        else
                            {
                            LOGWRITE(Format("%s: Disconnect IP: %s not found",
                                ARRAYOFCONST((clname, ip_str))),
                                LOGTASK_CLIENT_CTL, LOGL_ERROR);
                            }
                        }
                    else
                        {
                        LOGWRITE(Format("%s: Disconnect IP: bad request", ARRAYOFCONST((clname))),
                            LOGTASK_CLIENT_CTL, LOGL_ERROR);
                        }
                    ack(ack_flag);
                    }
                    break;

                //   IP    auto
                case CONTROL_COMMAND_IP_CONNECT_ALL_AUTO:
                    {
                    client->crate_service->PostControlMessage(LTR_SRV_autoconnect_ip);
                    LOGWRITE(Format("%s: Connect all IP (auto)", ARRAYOFCONST((clname))),
                        LOGTASK_CLIENT_CTL, LOGL_VERBOSE);
                    ack_flag = true;
                    ack(ack_flag);
                    }
                    break;

                //   IP 
                case CONTROL_COMMAND_IP_DISCONNECT_ALL:
                    {
                    client->crate_service->PostControlMessage(LTR_SRV_disconnect_all_ip);
                    LOGWRITE(Format("%s: Disconnect all IP", ARRAYOFCONST((clname))),
                        LOGTASK_CLIENT_CTL, LOGL_VERBOSE);
                    ack_flag = true;
                    ack(ack_flag);
                    }
                    break;

                case CONTROL_COMMAND_IP_SET_FLAGS:
                    {
                    struct //    (  )
                        {
                        DWORD ip_addr;     // IP- (host-endian)
                        DWORD flags;       //  
                        DWORD permanent;   //   ini 
                        }
                        params;
                    res = recv((BYTE*)&params, sizeof(params));
                    if (res == sizeof(params))
                        {
                        TAddrListNode node;
                        DWORD ip = htonl(params.ip_addr);
                        AnsiString ip_str = IPToStr(ip);
                        if (g_Conf.SetIPCrateFlags(IPCrateList, ip, CRATE_IP_FLAG__VALID_BITS_,
                            params.flags, params.permanent))
                            {
                            client->crate_service->PostDisplayMessage(LTR_DISP_ip_list_updated);
                            LOGWRITE(Format("%s: Set IP flags to %0.8X for %s%s",
                                ARRAYOFCONST((clname, params.flags & CRATE_IP_FLAG__VALID_BITS_, ip_str,
                                PERMSTR(params.permanent)))),
                                LOGTASK_CLIENT_CTL, LOGL_VERBOSE);
                            ack_flag = true;
                            }
                        else
                            {
                            LOGWRITE(Format("%s: Set IP flags: %s not found",
                                ARRAYOFCONST((clname, ip_str))),
                                LOGTASK_CLIENT_CTL, LOGL_ERROR);
                            }
                        }
                    else
                        {
                        LOGWRITE(Format("%s: Set IP flags: bad request", ARRAYOFCONST((clname))),
                            LOGTASK_CLIENT_CTL, LOGL_ERROR);
                        }
                    ack(ack_flag);
                    }
                    break;

                //    IP-
                case CONTROL_COMMAND_IP_GET_DISCOVERY_MODE:
                    {
                    struct //    (  )
                        {
                        DWORD enabled;
                        DWORD autoconnect;
                        }
                        reply;
                    reply.enabled = (g_Conf.EnableAutoDiscovery() ? 1 : 0);
                    reply.autoconnect = (g_Conf.AutoConnectDynamic() ? 1 : 0);
                    ack(true);
                    send((BYTE*)&reply, sizeof(reply));
                    }
                    break;

                //    IP-
                case CONTROL_COMMAND_IP_SET_DISCOVERY_MODE:
                    {
                    struct //    (  )
                        {
                        DWORD enabled;
                        DWORD autoconnect;
                        DWORD permanent;   //   ini 
                        }
                        params;
                    res = recv((BYTE*)&params, sizeof(params));
                    if (res == sizeof(params))
                        {
                        bool bEnable = (params.enabled != 0);
                        bool bAuto = (params.autoconnect != 0);
                        LOGWRITE(Format("%s: Set crate discovery mode (%s, autoconnect %s)%s",
                            ARRAYOFCONST((clname, (bEnable ? "enabled" : "disabled"), (bAuto ? "on" : "off"),
                            PERMSTR(params.permanent)))),
                            LOGTASK_CLIENT_CTL, LOGL_VERBOSE);
                        g_Conf.AutoConnectDynamic(bAuto, params.permanent);
                        if (g_Conf.EnableAutoDiscovery(bEnable, params.permanent) != bEnable)
                            { //   
                            client->crate_service->PostControlMessage(LTR_SRV_autodiscovery_changed);
                            }
                        client->crate_service->PostDisplayMessage(LTR_DISP_gui_vars_updated);
                        ack_flag = true;
                        }
                    else
                        {
                        LOGWRITE(Format("%s: Set crate discovery mode: bad request", ARRAYOFCONST((clname))),
                            LOGTASK_CLIENT_CTL, LOGL_ERROR);
                        }
                    ack(ack_flag);
                    }
                    break;

                //    
                case CONTROL_COMMAND_GET_SERVER_PROCESS_PRIORITY:
                    {
                    DWORD prioclass = GetPriorityClass(GetCurrentProcess());
                    ack(true);
                    send((BYTE*)&prioclass, sizeof(prioclass));
                    }
                    break;

                //    
                case CONTROL_COMMAND_SET_SERVER_PROCESS_PRIORITY:
                    {
                    DWORD prioclass;
                    ack(true); //   .   .
                    res = recv((BYTE*)&prioclass, sizeof(prioclass));
                    if (res == sizeof(prioclass))
                        {
                        LOGWRITE(Format("%s: Set server priority to 0x%0.8X",
                            ARRAYOFCONST((clname, prioclass))),
                            LOGTASK_CLIENT_CTL, LOGL_DEBUG1);
                        SetPriorityClass(GetCurrentProcess(), prioclass);
                        client->crate_service->PostDisplayMessage(LTR_DISP_gui_vars_updated);
                        ack_flag = true;
                        }
                    else
                        {
                        LOGWRITE(Format("%s: Set server priority: bad request", ARRAYOFCONST((clname))),
                            LOGTASK_CLIENT_CTL, LOGL_ERROR);
                        }
                    ack(ack_flag);
                    }
                    break;

                //   
                case CONTROL_COMMAND_GET_SERVER_VERSION:
                    {
                    ack(true);
                    send((BYTE*)&g_ServerVersion, sizeof(g_ServerVersion));
                    }
                    break;

                //   
                case CONTROL_COMMAND_GET_LOG_LEVEL:
                    {
                    LONG loglevel = g_Conf.LogLevel();
                    ack(true);
                    send((BYTE*)&loglevel, sizeof(loglevel));
                    }
                    break;

                //   
                case CONTROL_COMMAND_SET_LOG_LEVEL:
                    {
                    struct //    (  )
                        {
                        LONG loglevel;
                        DWORD permanent;   //   ini 
                        }
                        params;
                    res = recv((BYTE*)&params, sizeof(params));
                    if (res == sizeof(params))
                        {
                        int oldlevel = g_Conf.LogLevel(params.loglevel, params.permanent);
                        if (oldlevel != params.loglevel)
                            { // Log level changed
                            LOGWRITE(Format("%s: Log level changed from %d to %d%s",
                                ARRAYOFCONST((clname, oldlevel, params.loglevel,
                                PERMSTR(params.permanent)))),
                                LOGTASK_CLIENT_CTL, LOGL_CRITICAL);
                            }
                        g_Log.SetLevel(g_Conf.LogLevel());
                        client->crate_service->PostDisplayMessage(LTR_DISP_gui_vars_updated);
                        ack_flag = true;
                        }
                    else
                        {
                        LOGWRITE(Format("%s: Set log level: bad request", ARRAYOFCONST((clname))),
                            LOGTASK_CLIENT_CTL, LOGL_ERROR);
                        }
                    ack(ack_flag);
                    }
                    break;

                //  
                case CONTROL_COMMAND_RESTART_SERVER:
                    LOGWRITE(Format("%s: RESTART SERVER", ARRAYOFCONST((clname))),
                        LOGTASK_CLIENT_CTL, LOGL_CRITICAL);
                    ack(true);
                    client->crate_service->PostDisplayMessage(LTR_DISP_request_restart);
                    break;

                //  
                case CONTROL_COMMAND_SHUTDOWN_SERVER:
                    LOGWRITE(Format("%s: SHUTDOWN SERVER", ARRAYOFCONST((clname))),
                        LOGTASK_CLIENT_CTL, LOGL_CRITICAL);
                    ack(true);
                    client->crate_service->PostDisplayMessage(LTR_DISP_request_shutdown);
                    break;

                default:
                    ack(false);
                    break;
                } /* switch (CommandCode) */
            }
        } /* while */

    if (pRawDataBuffer)
        delete[] pRawDataBuffer;

    if (pcrate && (client->flags & CC_RAW_DATA_FLAG))
        pcrate->ReadRawData = false;

    //  
    client->Close();
    //    
    LOGWRITE(Format("%s: Connection stopped", ARRAYOFCONST((clname))), LOGTASK_CLIENT_CTL, LOGL_INFO);
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
__fastcall TReadClientThread::TReadClientThread(TClient *_client, bool CreateSuspended)
    : TThread(true), client(_client)
    {
    Priority = tpHigher; //tpNormal;
    FreeOnTerminate = false;
    if (!CreateSuspended) Resume();
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
void __fastcall TReadClientThread::Execute()
    {
    WSABUF wb;
    WSAOVERLAPPED ov;
    HANDLE evt[2];
    bool ok;

    if (!client->pcrate || !client->pmodule)
        {
        client->Close();
        return;
        }

    const char* clname = client->clname;

    ZeroMemory(&ov, sizeof(ov));
    ov.hEvent = WSACreateEvent();

    client->read_rec.buf = Buf;
    client->read_rec.buf_len = sizeof(Buf) / sizeof(Buf[0]);
    ok = true;

    LOGWRITE(Format("%s: starting read thread", ARRAYOFCONST((clname))), LOGTASK_CLIENT_READ, LOGL_DEBUG1);

#if DEBUG_PROFILER
    Prf::Add _(5, true);
#endif
    while (ok && !Terminated && !client->IsTerminating())
        {
        DWORD WordsReady = 0;
        //     
        switch (client->pmodule->ClientGet(client->read_rec, WordsReady))
            {
            case TCrateModule::eOk:
                break;
            case TCrateModule::eOverflow:
                LOGWRITE(Format("%s: receive buffer overflow", ARRAYOFCONST((clname))),
                    LOGTASK_CLIENT_READ, LOGL_WARNING);
                Buf[0] = LTR_SIGNAL_RBUF_OVF;
                WordsReady = 1;
                break;
            case TCrateModule::eNotPresent:
                LOGWRITE(Format("%s: module taken out", ARRAYOFCONST((clname))),
                    LOGTASK_CLIENT_READ, LOGL_DEBUG1);
                ok = false;
                break;
            default:
                LOGWRITE(Format("%s: module error", ARRAYOFCONST((clname))),
                    LOGTASK_CLIENT_READ, LOGL_ERROR);
                ok = false;
                break;
            }
        if (!ok) break;
        if (WordsReady > 0)
            {
            wb.len = WordsReady * sizeof(DWORD);
            wb.buf = (BYTE*)Buf;
            DEBUGDUMP(wb.buf, wb.len, Format("to client %s", ARRAYOFCONST((clname))).c_str());
            DWORD xfer_size = 0;
            DWORD flag = 0;
            int res = WSASend(client->socket, &wb, 1, &xfer_size, 0, &ov, NULL);
            ok = (res != SOCKET_ERROR);
            if (!ok)
                {
                int err = WSAGetLastError();
                if (err == WSA_IO_PENDING)
                    {
                    evt[0] = ov.hEvent;
                    evt[1] = client->hTermEvent;
                    _P;
                    DWORD wait_res = WaitForMultipleObjects(2, evt, FALSE, INFINITE);
                    _R;
                    switch (wait_res)
                        {
                        case WSA_WAIT_EVENT_0:  //  
                            ok = WSAGetOverlappedResult(client->socket, &ov, &xfer_size, FALSE, &flag);
                            if (!ok)
                                { //    
                                err = WSAGetLastError();
                                LOGWRITE(Format("%s: async send failed (error %d)",
                                    ARRAYOFCONST((clname, err))),
                                    LOGTASK_CLIENT_READ, LOGL_ERROR);
                                }
                            break;
                        case WSA_WAIT_EVENT_0 + 1:  //  
                            CancelIo((HANDLE)client->socket);
                            WSAGetOverlappedResult(client->socket, &ov, &xfer_size, TRUE, &flag);
                            LOGWRITE(Format("%s: send aborted by client shutdown", ARRAYOFCONST((clname))),
                                LOGTASK_CLIENT_READ, LOGL_DEBUG3);
                            break;
                        default:                // 
                            CancelIo((HANDLE)client->socket);
                            WSAGetOverlappedResult(client->socket, &ov, &xfer_size, TRUE, &flag);
                            LOGWRITE(Format("%s: send failed (wait error)", ARRAYOFCONST((clname))),
                                LOGTASK_CLIENT_READ, LOGL_ERROR);
                            break;
                        } /* switch */
                    }
                else
                    { //   , 
                    LOGWRITE(Format("%s: send failed (error %d)",
                        ARRAYOFCONST((clname, err))),
                        LOGTASK_CLIENT_READ, LOGL_DEBUG1);
                    }
                } /* if !ok */
            } /* if WordsReady > 0 */
        else
            { //  ,   
#if !PER_CLIENT_CRATE_EVENTS
            evt[0] = client->pmodule->hCratePutEvent;
#else
            evt[0] = client->hModulePutEvent;
#endif
            evt[1] = client->hTermEvent;
            DWORD wait_res;
            ok = false;
            do
                {
                if (!client->pmodule->Present())
                    {
                    LOGWRITE(Format("%s: module taken out", ARRAYOFCONST((clname))),
                        LOGTASK_CLIENT_READ, LOGL_DEBUG1);
                    break;
                    }
                _P;
                wait_res = WaitForMultipleObjects(2, evt, FALSE, TIMEOUT_WAIT_CRATEPUT);
                _R;
                switch (wait_res)
                    {
                    case WAIT_OBJECT_0:
                        ok = true;
                        break;
                    case WAIT_OBJECT_0 + 1:
                        LOGWRITE(Format("%s: wait aborted by client shutdown", ARRAYOFCONST((clname))),
                            LOGTASK_CLIENT_READ, LOGL_DEBUG3);
                        break;
                    case WAIT_FAILED:
                        LOGWRITE(Format("%s: wait error", ARRAYOFCONST((clname))),
                            LOGTASK_CLIENT_READ, LOGL_ERROR);
                        break;
                    }
                }
            while (wait_res == WAIT_TIMEOUT);
            }
        } /* while */

    _P;
    LOGWRITE(Format("%s: stopping read thread", ARRAYOFCONST((clname))), LOGTASK_CLIENT_READ, LOGL_DEBUG1);

    WSACloseEvent(ov.hEvent);
    client->RequestTermination();
    //    
    if (client->send_buf_thread)
        client->send_buf_thread->WaitFor();

    //  
    client->Close();
    //    
    LOGWRITE(Format("%s: Connection stopped", ARRAYOFCONST((clname))), LOGTASK_CLIENT_CTL, LOGL_INFO);
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
__fastcall TWriteClientThread::TWriteClientThread(TClient *_client, bool CreateSuspended)
    : TThread(true), client(_client)
    {
    Priority = tpHigher; //tpNormal;
    FreeOnTerminate = false;
    if (!CreateSuspended) Resume();
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
#define CLIENT_RECV_SMALL_THRESHOLD 128
void __fastcall TWriteClientThread::Execute()
    {
    WSABUF wb;
    WSAOVERLAPPED ov;
    HANDLE evt[2];
    DWORD RxOffset = 0; //  0..3     DWORD
    bool ok = true;

    if (!client->pcrate || !client->pmodule)
        return;

    const char* clname = client->clname;

    ZeroMemory(&ov, sizeof(ov));
    ov.hEvent = WSACreateEvent();

    LOGWRITE(Format("%s: starting write thread", ARRAYOFCONST((clname))), LOGTASK_CLIENT_WRITE, LOGL_DEBUG1);

#if DEBUG_PROFILER
    Prf::Add _(4, true);
#endif    
    while (ok && !Terminated && !client->IsTerminating())
        {
        //     
        wb.buf = (BYTE*)Buf + RxOffset;
        wb.len = sizeof(Buf) - RxOffset;
        DWORD xfer_size = 0;
        DWORD flag = 0;
        int res = WSARecv(client->socket, &wb, 1, &xfer_size, &flag, &ov, NULL);
        ok = (res != SOCKET_ERROR);
        if (!ok)
            {
            int err = WSAGetLastError();
            if (err == WSA_IO_PENDING)
                {
                evt[0] = ov.hEvent;
                evt[1] = client->hTermEvent;
                _P;
                DWORD wait_res = WaitForMultipleObjects(2, evt, FALSE, INFINITE);
                _R;
                switch (wait_res)
                    {
                    case WSA_WAIT_EVENT_0:  //  
                        ok = WSAGetOverlappedResult(client->socket, &ov, &xfer_size, FALSE, &flag);
                        if (!ok)
                            { //    
                            err = WSAGetLastError();
                            LOGWRITE(Format("%s: async receive failed (error %d)",
                                ARRAYOFCONST((clname, err))),
                                LOGTASK_CLIENT_WRITE, LOGL_ERROR);
                            }
                        break;
                    case WSA_WAIT_EVENT_0 + 1:  //  
                        CancelIo((HANDLE)client->socket);
                        WSAGetOverlappedResult(client->socket, &ov, &xfer_size, TRUE, &flag);
                        LOGWRITE(Format("%s: receive aborted by client shutdown", ARRAYOFCONST((clname))),
                            LOGTASK_CLIENT_WRITE, LOGL_DEBUG3);
                        break;
                    default:                // 
                        CancelIo((HANDLE)client->socket);
                        WSAGetOverlappedResult(client->socket, &ov, &xfer_size, TRUE, &flag);
                        LOGWRITE(Format("%s: receive failed (wait error)", ARRAYOFCONST((clname))),
                            LOGTASK_CLIENT_WRITE, LOGL_ERROR);
                        break;
                    } /* switch */
                }
            else
                { //   , 
                LOGWRITE(Format("%s: receive failed (error %d)",
                    ARRAYOFCONST((clname, err))),
                    LOGTASK_CLIENT_WRITE, LOGL_DEBUG1);
                }
            } /* if !ok */

        if (ok && (xfer_size == 0))
            {
            ok = false;
            LOGWRITE(Format("%s: received end-of-connection", ARRAYOFCONST((clname))),
                LOGTASK_CLIENT_WRITE, LOGL_DEBUG1);
            }
        if (!ok) break;

        // :   , xfer_size =  
        //            
        xfer_size += RxOffset;
        DWORD WordsReady = xfer_size / sizeof(DWORD);
        RxOffset = xfer_size % sizeof(DWORD); //  
        DEBUGDUMP(Buf, xfer_size, Format("from client %s", ARRAYOFCONST((clname))).c_str());
        if (WordsReady > 0)
            {
            /* !!!  :
                  
            (  ),      -  .
                  ,    
             .
            */
            if (WordsReady < CLIENT_RECV_SMALL_THRESHOLD)
                Sleep(2);

            DWORD and_mask, or_mask;
            //     
            if (!(client->flags & CC_DEBUG_FLAG))
                {
                and_mask = 0xFFFF80FF;
                or_mask = ((DWORD)client->crate_channel - 1) << 8;
                }
            else
                {
                and_mask = 0xFFFFFFFF;
                or_mask = 0;
                }
            //    
            DWORD SendPos = 0, sent;
            while (ok && (SendPos < WordsReady))
                {
                //while (!client->pmodule->SendBufEmpty()) Sleep(5);

                switch (client->pmodule->ClientPut(&Buf[SendPos], WordsReady - SendPos,
                                          sent, and_mask, or_mask))
                    {
                    case TCrateModule::eOk:
                        // ,     
                        SetEvent(client->pcrate->hClientPutEvent);
                        SendPos += sent;
                        break;
                    case TCrateModule::eWrongOp:
                        //      --  
                        SendPos = WordsReady;
                        break;
                    case TCrateModule::eNotPresent:
                        LOGWRITE(Format("%s: module taken out", ARRAYOFCONST((clname))),
                            LOGTASK_CLIENT_WRITE, LOGL_WARNING);
                        ok = false;
                        break;
                    default:
                        ok = false;
                        break;
                    }
                if (!ok) break;
                //    ,     
                if (SendPos < WordsReady)
                    {
#if !PER_CLIENT_CRATE_EVENTS
                    evt[0] = client->pmodule->hCrateGetEvent;
#else
                    evt[0] = client->hModuleGetEvent;
#endif
                    evt[1] = client->hTermEvent;
                    DWORD wait_res;
                    ok = false;
                    do
                        {
                        if (!client->pmodule->Present())
                            {
                            LOGWRITE(Format("%s: module taken out", ARRAYOFCONST((clname))),
                                LOGTASK_CLIENT_WRITE, LOGL_DEBUG1);
                            break;
                            }
                        _P;
                        wait_res = WaitForMultipleObjects(2, evt, FALSE, TIMEOUT_WAIT_CRATEGET);
                        _R;
                        switch (wait_res)
                            {
                            case WAIT_OBJECT_0:
                                ok = true;
                                break;
                            case WAIT_OBJECT_0 + 1:
                                LOGWRITE(Format("%s: wait aborted by client shutdown", ARRAYOFCONST((clname))),
                                    LOGTASK_CLIENT_WRITE, LOGL_DEBUG3);
                                break;
                            case WAIT_FAILED:
                                LOGWRITE(Format("%s: wait error", ARRAYOFCONST((clname))),
                                    LOGTASK_CLIENT_WRITE, LOGL_ERROR);
                                break;
                            }
                        }
                    while (wait_res == WAIT_TIMEOUT);
                    }
                } /* while ok && sendpos */

            //  ,     ,  
            if (RxOffset > 0)
                memcpy((BYTE*)Buf, (BYTE*)Buf + xfer_size - RxOffset, RxOffset);
            } /* if WordsReady > 0 */

        } /* while ok & ... */
    _P;

    LOGWRITE(Format("%s: stopping write thread", ARRAYOFCONST((clname))), LOGTASK_CLIENT_WRITE, LOGL_DEBUG1);

    WSACloseEvent(ov.hEvent);
    client->RequestTermination();
    }
/*------------------------------------------------------------------------------------------------*/

