/*================================================================================================*
 *     ltrserver.
 *  INI-      thread-safe  
 *  .
 *================================================================================================*/

#include <vcl.h>
#include <stdio.h>
#pragma hdrstop

#include "common.h"
#include "LtrConfig.h"

/*================================================================================================*/
/*  INI- ( ) */
const char INI_FILE_NAME[]              = "ltrserver.ini";
/*    INI- */
const char INI_SEC_CONFIG[]             = "LTR_SERVER_CONFIG";
const char INI_SEC_LTR_DEFAULT[]        = "LTR010_DEFAULT";
const char INI_SEC_IP_CRATES[]          = "LTR_IP_CRATES";
const char INI_SEC_LOG[]                = "LOG_SETTINGS";
const char INI_SEC_VISUAL[]             = "WINDOW_STATE";
const char INI_SEC_ADV_IP_SET[]         = "ADVANCED_IP_SETTINGS";
/*    INI- */
const char INI_KEY_PRIORITY[]           = "server_priority";
const char INI_KEY_MINIMIZE2TRAY[]      = "minimize_to_tray";
const char INI_KEY_STARTMINIMIZED[]     = "start_minimized";
const char INI_KEY_AUTOCONNECT[]        = "autoconnect_dynamic";
const char INI_KEY_AUTODISCOVER[]       = "enable_discovery";
const char INI_KEY_AUTODISCOVERSUPP[]   = "udp_discovery_support";
const char INI_KEY_FPGA_TTF[]           = "fpga_ttf";
const char INI_KEY_LOGNAME[]            = "logfile";
const char INI_KEY_LOGLEVEL[]           = "loglevel";
const char INI_KEY_LANG[]               = "language";
const char INI_KEY_ACTIVE_TAB[]         = "active_tab";
const char INI_KEY_MAXIMIZED[]          = "maximized";
const char INI_KEY_PLACEMENT[]          = "placement";
const char INI_KEY_LISTENADDR[]         = "listen_addr";
const char INI_KEY_LISTENPORT[]         = "listen_port";
const char INI_KEY_MULTICONNECT[]       = "multiconnect";
const char INI_KEY_LONGTSTAMP[]         = "enable_long_tstamp";
const char INI_KEY_CRATECONNTIMEOUT[]   = "crate_connect_timeout";
const char INI_KEY_CRATECMDTIMEOUT[]    = "crate_command_timeout";
const char INI_KEY_CRATEPINGTIMEOUT[]   = "crate_ping_timeout";
const char INI_KEY_CRATEIDLETIMEOUT[]   = "crate_idle_timeout";
const char INI_KEY_CRATERECONNDELAY[]   = "crate_reconnect_delay";
const char INI_KEY_CRATECONNATTEMPTS[]  = "crate_connect_attempts";
const char INI_KEY_CRATEPINGATTEMPTS[]  = "crate_ping_attempts";
/*================================================================================================*/

/*================================================================================================*/
/*------------------------------------------------------------------------------------------------*/
void TLtrConfig::Open()
    {
    /*  INI- */
    ini.Open(INI_FILE_NAME);

    /*   */
    language =
        ini.ReadString(INI_SEC_CONFIG, INI_KEY_LANG, "auto", true);

    /*   */
    log_file_name =
        ini.ReadString(INI_SEC_LOG, INI_KEY_LOGNAME, "ltrserver.log", true);
    int LL =
        ini.ReadInteger(INI_SEC_LOG, AnsiString(INI_KEY_LOGLEVEL), LOGL_INFO, true);
    if (LL < 0)
        LL = 0;
    if (LL > max_log_level)
        LL = max_log_level;
    log_level = LL;

    /*   */
    server_priority =
        ini.ReadInteger(INI_SEC_CONFIG, INI_KEY_PRIORITY, NORMAL_PRIORITY_CLASS, true);
    minimize_to_tray =
        ini.ReadBool(INI_SEC_CONFIG, INI_KEY_MINIMIZE2TRAY, true, true);
    start_minimized =
        ini.ReadBool(INI_SEC_CONFIG, INI_KEY_STARTMINIMIZED, false, true);
    auto_discovery_support =
        ini.ReadBool(INI_SEC_CONFIG, INI_KEY_AUTODISCOVERSUPP, false, false);
    auto_connect_dynamic = auto_discovery_support.get() &&
        ini.ReadBool(INI_SEC_CONFIG, INI_KEY_AUTOCONNECT, true, true);
    enable_auto_discovery = auto_discovery_support.get() &&
        ini.ReadBool(INI_SEC_CONFIG, INI_KEY_AUTODISCOVER, true, true);
    default_ttf = TLogfile::RelativeFileName(
        ini.ReadString(INI_SEC_LTR_DEFAULT, INI_KEY_FPGA_TTF, "ltr010.ttf", true)
        );
    listen_addr = // If invalid, returns 0 = 0.0.0.0 = INADDR_ANY
        StrToIP(ini.ReadString(INI_SEC_CONFIG, INI_KEY_LISTENADDR, "0.0.0.0", false));
    listen_port =
        ini.ReadInteger(INI_SEC_CONFIG, INI_KEY_LISTENPORT, SPORT_DEFAULT, false);
    multiconnect_to_module =
        ini.ReadBool(INI_SEC_CONFIG, INI_KEY_MULTICONNECT, false, false);
    enable_long_tstamp =
        ini.ReadBool(INI_SEC_CONFIG, INI_KEY_LONGTSTAMP, false, true);
    crate_connect_timeout =
        std::max(0, ini.ReadInteger(INI_SEC_ADV_IP_SET, INI_KEY_CRATECONNTIMEOUT, LTR_IP_DEFAULT_CONN_TIMEOUT, true));
    crate_command_timeout =
        std::max(0, ini.ReadInteger(INI_SEC_ADV_IP_SET, INI_KEY_CRATECMDTIMEOUT, LTR_IP_DEFAULT_IO_TIMEOUT, true));
    crate_ping_timeout =
        std::max(0, ini.ReadInteger(INI_SEC_ADV_IP_SET, INI_KEY_CRATEPINGTIMEOUT, LTR_IP_DEFAULT_IO_TIMEOUT, true));
    crate_idle_timeout =
        std::max(0, ini.ReadInteger(INI_SEC_ADV_IP_SET, INI_KEY_CRATEIDLETIMEOUT, 5000, true));
    crate_reconnect_delay =
        std::max(0, ini.ReadInteger(INI_SEC_ADV_IP_SET, INI_KEY_CRATERECONNDELAY, 2000, true));
    crate_connect_attempts =
        std::max(0, ini.ReadInteger(INI_SEC_ADV_IP_SET, INI_KEY_CRATECONNATTEMPTS, 10, true));
    crate_ping_attempts =
        std::max(0, ini.ReadInteger(INI_SEC_ADV_IP_SET, INI_KEY_CRATEPINGATTEMPTS, 3, true));
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
void TLtrConfig::Close()
    {
    ini.Close();
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
DWORD TLtrConfig::ServerPriority(DWORD dwNewVal, bool bSave)
    {
    if (bSave)
        ini.WriteInteger(INI_SEC_CONFIG, INI_KEY_PRIORITY, dwNewVal);
    return server_priority.xchg(dwNewVal);
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
bool TLtrConfig::MinimizeToTray(bool bNewVal, bool bSave)
    {
    if (bSave)
        ini.WriteBool(INI_SEC_CONFIG, INI_KEY_MINIMIZE2TRAY, bNewVal);
    return minimize_to_tray.xchg(bNewVal);
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
bool TLtrConfig::StartMinimized(bool bNewVal, bool bSave)
    {
    if (bSave)
        ini.WriteBool(INI_SEC_CONFIG, INI_KEY_STARTMINIMIZED, bNewVal);
    return start_minimized.xchg(bNewVal);
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
bool TLtrConfig::AutoConnectDynamic(bool bNewVal, bool bSave)
    {
    if (!auto_discovery_support.get()) return false;
    if (bSave)
        ini.WriteBool(INI_SEC_CONFIG, INI_KEY_AUTOCONNECT, bNewVal);
    return auto_connect_dynamic.xchg(bNewVal);
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
bool TLtrConfig::EnableAutoDiscovery(bool bNewVal, bool bSave)
    {
    if (!auto_discovery_support.get()) return false;
    if (bSave)
        ini.WriteBool(INI_SEC_CONFIG, INI_KEY_AUTODISCOVER, bNewVal);
    return enable_auto_discovery.xchg(bNewVal);
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
int TLtrConfig::LogLevel(int iNewVal, bool bSave)
    {
    if (iNewVal < 0)
        iNewVal = 0;
    if (iNewVal > max_log_level)
        iNewVal = max_log_level;
    if (bSave)
        ini.WriteInteger(INI_SEC_LOG, INI_KEY_LOGLEVEL, iNewVal);
    return log_level.xchg(iNewVal);
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
void TLtrConfig::Language(const AnsiString& sNewVal, bool bSave)
    {
    if (bSave)
        ini.WriteString(INI_SEC_CONFIG, INI_KEY_LANG, sNewVal);
    language = sNewVal;
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
DWORD TLtrConfig::ListenAddr(DWORD dwNewVal, bool bSave)
    {
    if (bSave)
        ini.WriteString(INI_SEC_CONFIG, INI_KEY_LISTENADDR, IPToStr(dwNewVal));
    return listen_addr.xchg(dwNewVal);
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
WORD TLtrConfig::ListenPort(WORD wNewVal, bool bSave)
    {
    if (bSave)
        ini.WriteInteger(INI_SEC_CONFIG, INI_KEY_LISTENPORT, wNewVal);
    return listen_port.xchg(wNewVal);
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
bool TLtrConfig::Multiconnect(bool bNewVal, bool bSave)
    {
    if (bSave)
        ini.WriteBool(INI_SEC_CONFIG, INI_KEY_MULTICONNECT, bNewVal);
    return multiconnect_to_module.xchg(bNewVal);
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
bool TLtrConfig::LongTStampEnabled(bool bNewVal, bool bSave)
    {
    if (bSave)
        ini.WriteBool(INI_SEC_CONFIG, INI_KEY_LONGTSTAMP, bNewVal);
    return enable_long_tstamp.xchg(bNewVal);
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
DWORD TLtrConfig::CrateConnectTimeout(DWORD dwNewVal, bool bSave)
    {
    if (bSave)
        ini.WriteInteger(INI_SEC_ADV_IP_SET, INI_KEY_CRATECONNTIMEOUT, dwNewVal);
    return crate_connect_timeout.xchg(dwNewVal);
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
DWORD TLtrConfig::CrateCommandTimeout(DWORD dwNewVal, bool bSave)
    {
    if (bSave)
        ini.WriteInteger(INI_SEC_ADV_IP_SET, INI_KEY_CRATECMDTIMEOUT, dwNewVal);
    return crate_command_timeout.xchg(dwNewVal);
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
DWORD TLtrConfig::CratePingTimeout(DWORD dwNewVal, bool bSave)
    {
    if (bSave)
        ini.WriteInteger(INI_SEC_ADV_IP_SET, INI_KEY_CRATEPINGTIMEOUT, dwNewVal);
    return crate_ping_timeout.xchg(dwNewVal);
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
DWORD TLtrConfig::CrateIdleTimeout(DWORD dwNewVal, bool bSave)
    {
    if (bSave)
        ini.WriteInteger(INI_SEC_ADV_IP_SET, INI_KEY_CRATEIDLETIMEOUT, dwNewVal);
    return crate_idle_timeout.xchg(dwNewVal);
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
DWORD TLtrConfig::CrateReconnectDelay(DWORD dwNewVal, bool bSave)
    {
    if (bSave)
        ini.WriteInteger(INI_SEC_ADV_IP_SET, INI_KEY_CRATERECONNDELAY, dwNewVal);
    return crate_reconnect_delay.xchg(dwNewVal);
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
int TLtrConfig::CrateConnectAttempts(int iNewVal, bool bSave)
    {
    if (iNewVal < 0) iNewVal = 0;
    if (bSave)
        ini.WriteInteger(INI_SEC_ADV_IP_SET, INI_KEY_CRATECONNATTEMPTS, iNewVal);
    return crate_connect_attempts.xchg(iNewVal);
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
int TLtrConfig::CratePingAttempts(int iNewVal, bool bSave)
    {
    if (iNewVal < 0) iNewVal = 0;
    if (bSave)
        ini.WriteInteger(INI_SEC_ADV_IP_SET, INI_KEY_CRATEPINGATTEMPTS, iNewVal);
    return crate_ping_attempts.xchg(iNewVal);
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
void TLtrConfig::LoadVisualState(VisualState& state)
    {
    TCSLock lock(cs_visual);
    AnsiString pos_str = ini.ReadString(INI_SEC_VISUAL, INI_KEY_PLACEMENT, "", false);
    state.valid = (sscanf(pos_str.c_str(), "%ld,%ld,%ld,%ld",
        &state.rect.left, &state.rect.top, &state.rect.right, &state.rect.bottom) == 4);
    state.maximized =
        ini.ReadBool(INI_SEC_VISUAL, INI_KEY_MAXIMIZED, false, false);
    state.active_tab =
        ini.ReadInteger(INI_SEC_VISUAL, INI_KEY_ACTIVE_TAB, 0, false);
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
void TLtrConfig::SaveVisualState(const VisualState& state)
    {
    TCSLock lock(cs_visual);
    ini.WriteString(INI_SEC_VISUAL, INI_KEY_PLACEMENT, Format("%d,%d,%d,%d", ARRAYOFCONST((
        state.rect.left,
        state.rect.top,
        state.rect.right,
        state.rect.bottom
        ))));
    ini.WriteBool(INI_SEC_VISUAL, INI_KEY_MAXIMIZED, state.maximized);
    ini.WriteInteger(INI_SEC_VISUAL, INI_KEY_ACTIVE_TAB, state.active_tab);
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
void TLtrConfig::LoadIPCrates(TUniqueAddrMultiList& list)
    {
    TCSLock lock(cs_ip);
    list.Clear();
    if (TStrings* KeyList = new TStringList)
        {
        AnsiString mode;
        ini.ReadSection(INI_SEC_IP_CRATES, KeyList);
        for (int i = 0; i < KeyList->Count; i++)
            {
            mode = ini.ReadString(INI_SEC_IP_CRATES, KeyList->Strings[i], "", false);
            if (mode.IsEmpty()) continue;
            DWORD ip = StrToIP(KeyList->Strings[i]);
            DWORD flags = ((mode.LowerCase() == "auto") ? CRATE_IP_FLAG_AUTOCONNECT : 0);
            list.AddUnique(IPLIST_STATIC, ip, flags & CRATE_IP_FLAG__VALID_BITS_);
            }
        delete KeyList;
        }
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
bool TLtrConfig::AddIPCrate(TUniqueAddrMultiList& list, DWORD ip, DWORD flags, bool bSave)
    {
    TCSLock lock(cs_ip);
    bool result = list.AddUnique(IPLIST_STATIC, ip, flags);
    if (result && bSave)
        ini.WriteString(INI_SEC_IP_CRATES, IPToStr(ip),
            (flags & CRATE_IP_FLAG_AUTOCONNECT) ? "auto" : "manual");
    return result;
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
bool TLtrConfig::ChangeIPCrate(TUniqueAddrMultiList& list, DWORD old_ip, DWORD new_ip,
    DWORD new_flags, bool bSave)
    {
    TCSLock lock(cs_ip);
    bool result = list.ReplaceUnique(IPLIST_STATIC, old_ip, new_ip, new_flags);
    if (result && bSave)
        {
        if (old_ip != new_ip)
            ini.DeleteKey(INI_SEC_IP_CRATES, IPToStr(old_ip));
        ini.WriteString(INI_SEC_IP_CRATES, IPToStr(new_ip),
            (new_flags & CRATE_IP_FLAG_AUTOCONNECT) ? "auto" : "manual");
        }
    return result;
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
bool TLtrConfig::SetIPCrateFlags(TUniqueAddrMultiList& list, DWORD ip, DWORD mask, DWORD bits,
    bool bSave)
    {
    TCSLock lock(cs_ip);
    DWORD newflags;
    int list_idx = list.SetFlags(ip, mask, bits, &newflags);
    if ((list_idx == IPLIST_STATIC) && bSave)
        ini.WriteString(INI_SEC_IP_CRATES, IPToStr(ip),
            (newflags & CRATE_IP_FLAG_AUTOCONNECT) ? "auto" : "manual");
    return (list_idx >= 0);
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
bool TLtrConfig::DeleteIPCrate(TUniqueAddrMultiList& list, DWORD ip, bool bSave)
    {
    TCSLock lock(cs_ip);
    if (list.Delete(IPLIST_STATIC, ip))
        {
        if (bSave)
            ini.DeleteKey(INI_SEC_IP_CRATES, IPToStr(ip));
        return true;
        }
    else
        {
        return false;
        }
    }
/*------------------------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------------------------*/
AnsiString TLtrConfig::GetTTFFileName(const AnsiString& section)
    {
    return TLogfile::RelativeFileName(
        ini.ReadString(section, INI_KEY_FPGA_TTF, default_ttf.get(), false)
        );
    }
/*------------------------------------------------------------------------------------------------*/
/*================================================================================================*/
