#include "LtrClient.h"
#include "LtrServiceClients.h"
#include "LtrCrate.h"

//-----------------------------------------------------------------------------
__fastcall TLTRServiceClient::TLTRServiceClient(bool CreateSuspended, TLTRServiceCrate* pCrateService)
    : TThread(true),
    server(INVALID_SOCKET),
    crate_service(pCrateService),
    init_ok(false)
    {
    Priority = tpNormal;
    FreeOnTerminate = false;
    init_complete = CreateEvent(NULL, EVT_MANUAL_RESET, EVT_NONSIGNALED, NULL);
    if (!CreateSuspended) Resume();
    }
//-----------------------------------------------------------------------------
__fastcall TLTRServiceClient::~TLTRServiceClient(void)
    {
    CloseHandle(init_complete);
    }
//-----------------------------------------------------------------------------
void __fastcall TLTRServiceClient::Execute()
    {
    LOGWRITE("LTRServiceClient thread started.", LOGTASK_CLIENT_CTL, LOGL_VERBOSE);

    if (g_Log.CheckLevel(LOGL_VERBOSE))
        { //   , IP ,  
        char hostname[256];
        if (!gethostname(hostname, sizeof(hostname)))
            {
            AnsiString str = hostname;
            struct addrinfo *ai_list, hint;
            ZeroMemory(&hint, sizeof(hint));
            hint.ai_family = AF_INET;
            if (!getaddrinfo(hostname, NULL, &hint, &ai_list))
                {
                for (struct addrinfo *iter = ai_list; iter; iter = iter->ai_next)
                    {
                    if (iter->ai_addr->sa_family == AF_INET)
                        {
                        if (((sockaddr_in*)iter->ai_addr)->sin_addr.s_addr != 0x0100007F)
                            str += Format(" [%u.%u.%u.%u]", ARRAYOFCONST((
                                ((sockaddr_in*)iter->ai_addr)->sin_addr.S_un.S_un_b.s_b1,
                                ((sockaddr_in*)iter->ai_addr)->sin_addr.S_un.S_un_b.s_b2,
                                ((sockaddr_in*)iter->ai_addr)->sin_addr.S_un.S_un_b.s_b3,
                                ((sockaddr_in*)iter->ai_addr)->sin_addr.S_un.S_un_b.s_b4
                                )));
                        }
                    }
                freeaddrinfo(ai_list);
                }
            LOGWRITE("Server: " + str, LOGTASK_CLIENT_INIT, LOGL_VERBOSE);
            }

        DWORD nlen = sizeof(hostname);
        if (GetUserName(hostname, &nlen))
            LOGWRITE(AnsiString("User: ") + hostname, LOGTASK_CLIENT_INIT, LOGL_VERBOSE);
        }

    //  
    server = WSASocket(AF_INET, SOCK_STREAM, IPPROTO_TCP, NULL, 0, WSA_FLAG_OVERLAPPED);
    if (server != INVALID_SOCKET)
        {
        sockaddr_in server_addr;
        sockaddr peer_addr;
        server_addr.sin_family = AF_INET;
        server_addr.sin_port = htons(g_Conf.ListenPort());
        server_addr.sin_addr.s_addr = g_Conf.ListenAddr();
        peer_addr.sa_family = AF_INET;
        int peer_addr_len = sizeof(peer_addr);
        if (!bind(server, (sockaddr*)&server_addr, sizeof(server_addr)))
            {
            if (!listen(server, 16))
                {
                //   
                init_ok = true;
                SetEvent(init_complete);
                while (!Terminated)
                    {
                    SOCKET s = WSAAccept(server, &peer_addr, &peer_addr_len, NULL, NULL);
                    if ((s != INVALID_SOCKET) && (!Terminated))
                        {
                        DWORD peer_ip = (peer_addr.sa_family == AF_INET) ?
                            ((sockaddr_in*)&peer_addr)->sin_addr.s_addr : INADDR_NONE;
                        bool ok = false;

                        AnsiString connrefused = "Connection refused for client " + IPToStr(peer_ip) + ": ";

                        if (WaitForSingleObject(client_list_access, INFINITE)==WAIT_OBJECT_0)
                            {
                            //   
                            //      
                            int client_index = CLIENT_MAX;
                            for (int i = 0; i < CLIENT_MAX; i++)
                                {
                                if (client_list[i] == NULL)
                                    {
                                    client_index = i;
                                    }
                                else if (!client_list[i]->IsOpen())
                                    {
                                    delete client_list[i];
                                    client_list[i] = NULL;
                                    client_index = i;
                                    }
                                }
                            //   
                            if (client_index < CLIENT_MAX)
                                {
                                client_list[client_index] = new TClient(s, peer_ip, crate_service);
                                ok = true;
                                }
                            else
                                {
                                LOGWRITE(connrefused + "no empty slot for client.", LOGTASK_CLIENT_INIT, LOGL_ERROR);
                                }
                            ReleaseMutex(client_list_access);
                            } /* mutex */
                        else
                            {
                            LOGWRITE(connrefused + "cannot access client list.", LOGTASK_CLIENT_INIT, LOGL_ERROR);
                            }

                        if (!ok)
                            {
                            shutdown(s, SD_BOTH);
                            closesocket(s);
                            }
                        }
                    else
                        {
                        int error = WSAGetLastError();
                         /*   , accept()   WSAEINTR,   */
                        if (error != WSAEINTR)
                            LOGWRITE(Format("accept() returned error %d", ARRAYOFCONST((error))),
                                LOGTASK_CLIENT_INIT, LOGL_ERROR);
                        }
                    } /* while (!Terminated) */
                }
            else LOGWRITE(Format("Cannot listen for clients (error %d)", ARRAYOFCONST(((int)WSAGetLastError()))),
                 LOGTASK_CLIENT_INIT, LOGL_ERROR);
            }
        else LOGWRITE(Format("Cannot bind server socket to %s:%d (error %d). Check INI file.",
            ARRAYOFCONST((IPToStr(server_addr.sin_addr.s_addr), ntohs(server_addr.sin_port), (int)WSAGetLastError()))),
            LOGTASK_CLIENT_INIT, LOGL_ERROR);
        closesocket(server);
        }
    else LOGWRITE(Format("Cannot create server socket (error %d)", ARRAYOFCONST(((int)WSAGetLastError()))),
         LOGTASK_CLIENT_INIT, LOGL_ERROR);

    LOGWRITE("LTRServiceClient thread shutting down", LOGTASK_CLIENT_CTL, LOGL_DEBUG1);

    //   
    if (WaitForSingleObject(client_list_access, INFINITE) == WAIT_OBJECT_0)
        {
        //     ,   .
        for (int k = 0; k < CLIENT_MAX; k++)
            {
            if (client_list[k])
                client_list[k]->RequestTermination();
            }
        for (int k = 0; k < CLIENT_MAX; k++)
            {
            if (client_list[k])
                {
                delete client_list[k]; //    
                client_list[k] = NULL;
                }
            }
        ReleaseMutex(client_list_access);
        }
    else
        {
        LOGWRITE("Cannot access client list.", LOGTASK_CLIENT_CTL, LOGL_ERROR);
        }
    LOGWRITE("LTRServiceClient thread shut down.", LOGTASK_CLIENT_CTL, LOGL_VERBOSE);
    SetEvent(init_complete);
}
//-----------------------------------------------------------------------------
