#ifndef COMMONH
#define COMMONH

//  warning "Comparins signed and unsigned values"
#pragma warn -csu

#define FD_SETSIZE (1)
#include <winsock2.h>
#include <ws2tcpip.h>
#include <windows.h>
#include <stdio.h>
#include <classes.hpp>
#include <inifiles.hpp>
#include <algorithm>
#include <vector>
#include "ltr\\include\\ltrapi_ll.h"
#include "MessageQueue.h"
#include "addrlist.h"
#include "Logfile.h"
#include "debugdump.h"
#include "LtrConfig.h"
#include "usb_util.h"
//---------------------------------------------------------------------------

#ifdef _DEBUG
#define DEBUG_PROFILER      1
#else
#define DEBUG_PROFILER      0
#endif

#define USB_DETECT_ADVANCED 1

/* 23-Apr-2012 --  multiconnect   ,   "   "  
 "    " (CrateGetEvent, CratePutEvent)    ,  
  (..   auto reset).         
    */
#define PER_CLIENT_CRATE_EVENTS 1

#define EVT_MANUAL_RESET    TRUE
#define EVT_AUTO_RESET      FALSE
#define EVT_SIGNALED        TRUE
#define EVT_NONSIGNALED     FALSE

#define TIMEOUT_IP_SHORTER 10      /*     data connection (  ) */
#define TIMEOUT_IP_LONGER  200     /*    data connection,    */
#define TIMEOUT_WAIT_CLIENTPUT  10
#define TIMEOUT_WAIT_CRATEGET   50
#define TIMEOUT_WAIT_CRATEPUT   50

#define IP_LIST_MAX         256

/*     MODULE_MAX  .
 *  ,            */
#define MODULE_INDEX_RAW          (MODULE_MAX + 0)
#define MODULE_INDEX_USER         (MODULE_MAX + 1)
#define MODULE_ARRAYLEN           (MODULE_MAX + 2)

#define CLIENT_MAX          256

//---------------------------------------------------------------------------
#define CRATE_RR_N                  32                //       

//     
#define CRATE_RR_SIZE_MAX_DWORD     (128*1024l)

//#define MODULE_RECV_BUF_SIZE        (1*1024*1024l)   //     (DWORD[MODULE_RECV_BUF_SIZE])

#define CRATE_WR_SIZE_MAX_DWORD     (128*1024l)      //      (DWORD[CRATE_WR_SIZE_MAX])
//#define MODULE_SEND_BUF_SIZE        (128*1024l)      //     (DWORD[MODULE_SEND_BUF_SIZE])

//   IP  (  socket)
#define CRATE_IP_RECV_BUF_SIZE     (256 * 1024)   //      ( recv)
#define CRATE_IP_RECV_TCPBUF_SIZE  (CRATE_IP_RECV_BUF_SIZE * 2) //  ,  TCP 
#define CRATE_IP_SEND_BUF_SIZE     (256 * 1024)   //      ( send)
#define CRATE_IP_SEND_TCPBUF_SIZE  (CRATE_IP_SEND_BUF_SIZE * 2) //  ,  TCP 

//     (bootrom, application, auto)
#define CM_AUTO                 0       // (application_active) ? application : bootloader
#define CM_APP                  1       // application access
#define CM_BOOT                 2       // bootloader access

#define CRATE_FLAG_BOOT_LOADER  (1 << 0)
#define CRATE_FLAG_HW_ERROR     (1 << 1)

#define SCREEN_LOG_SIZE         (16000)

/*     */
#define LOGTASK_GLOBAL          "SERVER"
#define LOGTASK_CRATE_INIT      "CRATE_INIT"
#define LOGTASK_CRATE_CTL       "CRATE_CTL"
#define LOGTASK_CRATE_READ      "CRATE_READ"
#define LOGTASK_CRATE_WRITE     "CRATE_WRITE"
#define LOGTASK_CLIENT_INIT     "CLIENT_INIT"
#define LOGTASK_CLIENT_CTL      "CLIENT_CTL"
#define LOGTASK_CLIENT_READ     "CLIENT_READ"
#define LOGTASK_CLIENT_WRITE    "CLIENT_WRITE"

#define LOGWRITE(msg, task, level)  \
    do { if (g_Log.CheckLevel(level)) g_Log.Write((msg), (task), (level)); } while (0)

#define STRNCPYZ(dest, src) \
    do { strncpy((dest), (src), sizeof(dest)); (dest)[sizeof(dest) - 1] = '\0'; } while (0)

/*   */
enum enum_LogLevels
    {
    LOGL_CRITICAL = 0,
    LOGL_ERROR,
    LOGL_WARNING,
    LOGL_INFO,
    LOGL_VERBOSE,
    LOGL_DEBUG1,
    LOGL_DEBUG2,
    LOGL_DEBUG3
    };

/*       */
typedef enum enum_PrioRank
    {
    PRIORANK_IDLE           = 10,
    PRIORANK_BELOWNORMAL    = 20,
    PRIORANK_NORMAL         = 30,
    PRIORANK_ABOVENORMAL    = 40,
    PRIORANK_HIGH           = 50,
    PRIORANK_REALTIME       = 60
    }
    t_PrioRank;

//-----------------------------------------------------------------------------
//------------------------ Messages for LTRCrateServiceThread -----------------
//-----------------------------------------------------------------------------

/*   CrateServiceThread */
struct MsgControl
    {
    DWORD dwMessage;
    DWORD dwParam;
    MsgControl(DWORD code = 0, DWORD param = 0)
        : dwMessage(code), dwParam(param) { }
    void uniq() { }
    };

enum en_MsgControlCodes
    {
    LTR_SRV_rescan_usb = 0x1000,
#if USB_DETECT_ADVANCED
    LTR_SRV_insert_usb,                     /* strParam = device name */
    LTR_SRV_remove_usb,                     /* strParam = device name */
#endif
    LTR_SRV_autoconnect_ip,
    LTR_SRV_disconnect_all_ip,
    LTR_SRV_server_priority_changed,
    LTR_SRV_suspend,
    LTR_SRV_resume_suspend,
    LTR_SRV_bury_the_dead,                  /*  ,    */
    LTR_SRV_connect_crate,                  /* dwParam = ip */
    LTR_SRV_disconnect_crate,               /* dwParam = ip */
    LTR_SRV_autodiscovery_changed
    };

/*       */
struct MsgDisplay
    {
    DWORD dwMessage;
    MsgDisplay(DWORD code = 0) : dwMessage(code) { }
    void uniq() { }
    };

enum en_MsgDisplayCodes
    {
    LTR_DISP_crate_list_updated = 0x2000,
    LTR_DISP_ip_list_updated,
    LTR_DISP_gui_vars_updated,
    LTR_DISP_request_restart,
    LTR_DISP_request_shutdown
    };

/*    */
enum t_CrateState
    {
    CRATE_ST_NULL,          /*  CRATE_STATE()    */
    CRATE_ST_UNINIT,        /*        */
    CRATE_ST_CONNECTING,    /*     */
    CRATE_ST_ONLINE,        /*     */
    CRATE_ST_DEAD,          /*       */
    };

//---------------------------------------------------------------------------
class TLTRServiceCrate;
class TCrate;
class TCrateModule;
class TClient;

//---------------------------------------------------------------------------
extern HANDLE crate_list_access;
extern HANDLE client_list_access;
extern TCrate  *crate_list[CRATE_MAX];
extern TClient *client_list[CLIENT_MAX];
extern TUniqueAddrMultiList IPCrateList;
extern TLogfile g_Log;
extern TLtrConfig g_Conf;
extern DWORD g_ServerVersion;
//---------------------------------------------------------------------------
//  
DWORD StrToIP(AnsiString ip_str);
AnsiString IPToStr(DWORD ip);
t_PrioRank PriorityRank(DWORD priority_class);
//---------------------------------------------------------------------------

/*------------------------------------------------------------------------------------------------*/
template <class threadclass>
void StopThread(threadclass* &Thread)
    {
    if (Thread)
        {
        Thread->Terminate();
        Thread->WaitFor();
        delete Thread;
        Thread = NULL;
        }
    }
/*------------------------------------------------------------------------------------------------*/

#if DEBUG_PROFILER
struct Prf;
extern Prf g_Prf;
struct Prf
    {
    __int64 freq;
    __int64 counter[32];
    __int64 mark[32];
    Prf()
        { LARGE_INTEGER z; QueryPerformanceFrequency(&z);
        freq = z.QuadPart;
        memset(counter, 0, sizeof(counter)); }
    struct Add
        {
        int id;
        bool on;
        Pause()
            { if (on)
                {
                LARGE_INTEGER z; QueryPerformanceCounter(&z);
                g_Prf.counter[id] += (z.QuadPart - g_Prf.mark[id]);
                on = false;
            }   }
        Resume()
            { if (!on)
                {
                LARGE_INTEGER z; QueryPerformanceCounter(&z);
                g_Prf.mark[id] = z.QuadPart;
                on = true;
            }   }
        Add(int id_, bool start = true) : id(id_), on(false)
            { if (start) Resume(); }
        ~Add()
            { Pause(); }
        };
    };
#define _P  _.Pause()
#define _R  _.Resume()
#else
#define _P
#define _R
#endif

#endif
